import typing as t

from composio.client.enums.enum import Enum, EnumGenerator

from .base import TriggerData


_TRIGGER_CACHE: t.Dict[str, "Trigger"] = {}


class Trigger(Enum[TriggerData], metaclass=EnumGenerator):
    cache_folder = "triggers"
    cache = _TRIGGER_CACHE
    storage = TriggerData

    def load_from_runtime(self) -> t.Optional[TriggerData]:
        from composio.tools.base.abs import (  # pylint: disable=import-outside-toplevel
            trigger_registry,
        )

        for triggers in trigger_registry.values():
            if self.slug in triggers:
                self._data = TriggerData(
                    name=triggers[self.slug].name,
                    app=triggers[self.slug].tool,
                    path=self.storage_path,
                )
                return self._data

        return None

    def fetch_and_cache(self) -> t.Optional[TriggerData]:
        from composio.client import Composio  # pylint: disable=import-outside-toplevel
        from composio.client.endpoints import (  # pylint: disable=import-outside-toplevel
            v2,
        )

        client = Composio.get_latest()

        # TODO: client.triggers.endpoint is still v1, migrate that
        response = client.http.get(url=str(v2.triggers / self.slug)).json()
        # TOFIX: Return proper error code when of item is not found
        if "appName" not in response:
            return None

        return TriggerData(  # type: ignore
            name=response["enum"],
            app=response["appName"],
            path=self.storage_path,
        )

    @property
    def name(self) -> str:
        return self.load().name

    @property
    def app(self) -> str:
        # Normalize app name
        return self.load().app.upper()



import typing as t
import warnings

from composio.client.enums.base import ActionData, replacement_action_name
from composio.client.enums.enum import Enum, EnumGenerator
from composio.exceptions import ComposioSDKError


_ACTION_CACHE: t.Dict[str, "Action"] = {}


class Action(Enum[ActionData], metaclass=EnumGenerator):
    cache_folder = "actions"
    cache = _ACTION_CACHE
    storage = ActionData

    def load(self) -> ActionData:
        """Handle deprecated actions"""
        action_data = super().load()
        if action_data.replaced_by is not None:
            replacement_enum = Action(action_data.replaced_by)
            warnings.warn(
                f"{self!r} is deprecated and will be removed. "
                f"Use {replacement_enum!r} instead.",
                UserWarning,
            )
            return replacement_enum.load()

        return action_data

    def load_from_runtime(self) -> t.Optional[ActionData]:
        """Try to see if the action is a runtime action."""
        from composio.tools.base.abs import (  # pylint: disable=import-outside-toplevel
            action_registry,
        )

        for gid, actions in action_registry.items():
            if self.slug in actions:
                action = actions[self.slug]
                self._data = ActionData(
                    name=action.name,
                    app=action.tool,
                    tags=action.tags(),
                    no_auth=action.no_auth,
                    is_local=gid in ("runtime", "local"),
                    is_runtime=gid == "runtime",
                    path=self.storage_path,
                )
                return self._data

        return None

    def fetch_and_cache(self) -> t.Optional[ActionData]:
        from composio.client import Composio  # pylint: disable=import-outside-toplevel

        client = Composio.get_latest()
        request = client.http.get(url=str(client.actions.endpoint / self.slug))
        response = request.json()
        if isinstance(response, list):
            response, *_ = response

        if request.status_code == 404 or "Not Found" in response.get("message", ""):
            raise ComposioSDKError(
                message=(
                    f"No metadata found for enum `{self.slug}`, "
                    "You might be trying to use an app or action "
                    "that is deprecated."
                )
            )

        # TOFIX: Return proper error code when of item is not found
        if "appName" not in response:
            return None

        replaced_by = replacement_action_name(
            response["description"], response["appName"]
        )
        return ActionData(  # type: ignore
            name=response["name"],
            app=response["appName"],
            tags=response["tags"],
            no_auth=(
                client.http.get(url=str(client.apps.endpoint / response["appName"]))
                .json()
                .get("no_auth", False)
            ),
            is_local=False,
            is_runtime=False,
            shell=False,
            path=self.storage_path,
            replaced_by=replaced_by,
        )

    @property
    def name(self) -> str:
        """Action name."""
        return self.load().name

    @property
    def app(self) -> str:
        """App name where the actions belongs to."""
        # Normalize app name
        return self.load().app.upper()

    @property
    def tags(self) -> t.List[str]:
        """List of tags for action."""
        return self.load().tags

    @property
    def no_auth(self) -> bool:
        """If set `True` the action does not require authentication."""
        return self.load().no_auth

    @property
    def is_local(self) -> bool:
        """If set `True` the `app` is a local app."""
        return self.load().is_local

    @property
    def is_runtime(self) -> bool:
        """If set `True` the `app` is a runtime app."""
        return self.load().is_runtime



import typing as t

from composio.tools.base.local import LocalAction, LocalTool

from .actions import Calculator


class Mathematical(LocalTool, autoload=True):
    """Mathematical Tools for LLM"""

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/mathematical.png"

    @classmethod
    def actions(cls) -> list[t.Type[LocalAction]]:
        return [Calculator]



import os
import typing as t
from pathlib import Path

import typing_extensions as te

from composio.constants import LOCAL_CACHE_DIRECTORY
from composio.exceptions import ComposioSDKError
from composio.storage.base import LocalStorage

from .base import ActionData, EnumStringNotFound, SentinalObject


DataT = t.TypeVar("DataT", bound=LocalStorage)


NO_REMOTE_ENUM_FETCHING = (
    os.environ.get("COMPOSIO_NO_REMOTE_ENUM_FETCHING", "false") != "false"
)


class Enum(t.Generic[DataT]):
    cache_folder: str
    cache: t.Dict[str, "te.Self"]
    storage: t.Type[DataT]

    def __new__(cls, value: t.Union[str, te.Self, t.Type[SentinalObject]]) -> "te.Self":
        """Cache the enum singleton."""
        # No caching for runtime actions
        if hasattr(value, "sentinel"):  # TODO: get rid of SentinalObject
            return super().__new__(cls)

        # If we get an enum object, return it as is
        if isinstance(value, cls):
            return value

        # Because people will pass weird stuff into the constructor
        value = str(value)

        # Normalize slug
        value = value.upper()

        cached_enum = cls.cache.get(value)
        if cached_enum is not None:
            return cached_enum  # type: ignore[return-value]

        enum = super().__new__(cls)
        cls.cache[value] = enum
        return enum

    def __init__(self, value: t.Union[str, te.Self, t.Type[SentinalObject]]) -> None:
        if hasattr(self, "_data"):
            # Object was pulled from cache and is already initialized
            return

        self._data: t.Optional[DataT] = None

        # If we get an enum object, return it as is
        if isinstance(value, self.__class__):
            return

        # Runtime action handling
        if hasattr(value, "sentinel"):  # TODO: get rid of SentinalObject
            slug = value.enum  # type: ignore
            if not isinstance(slug, str):
                raise ComposioSDKError(f"Invalid enum type: {slug!r}, expected str")
        else:
            slug = str(value)

        # Normalize slug
        slug = slug.upper()
        self.slug = slug

    def __hash__(self) -> int:
        return hash(self.slug)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.slug}"

    def __str__(self) -> str:
        return self.slug

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Enum):
            return self.slug == other.slug

        if isinstance(other, str):
            return self.slug == other

        return False

    @classmethod
    def iter(cls) -> t.Iterator[str]:
        """Yield the enum names as strings."""
        # TODO: fetch trigger names from dedicated endpoint in the future
        path = LOCAL_CACHE_DIRECTORY / cls.cache_folder
        # If we try to fetch Actions.iter() with local caching disabled
        # for example, we'd get here.
        if not path.exists():
            # pylint: disable=import-outside-toplevel
            from composio.client import Composio

            # pylint: disable=import-outside-toplevel
            from composio.client.utils import check_cache_refresh

            check_cache_refresh(Composio.get_latest())
            if not path.exists():
                return

        yield from os.listdir(path)

    @classmethod
    def all(cls) -> t.Iterator[te.Self]:
        """Iterate over available object."""
        for app_name in cls.iter():
            yield cls(app_name)

    @property
    def storage_path(self) -> Path:
        return LOCAL_CACHE_DIRECTORY / self.cache_folder / self.slug

    def load(self) -> DataT:
        if self._data is not None:
            return self._data

        if self.storage_path.exists():
            data = self.storage.load(self.storage_path)
            # HACK: if 'replaced_by' field is not present, delete this cached file
            # as it is outdated.
            if isinstance(data, ActionData):
                if hasattr(data, "replaced_by"):
                    self._data = data  # type: ignore
                    return self._data  # type: ignore

                self.storage_path.unlink()

            self._data = data
            return self._data

        # Try to fetch from runtime
        runtime_data = self.load_from_runtime()
        if runtime_data is not None:
            self._data = runtime_data
            return self._data

        # Try to fetch from API, and cache it locally
        if not NO_REMOTE_ENUM_FETCHING:
            remote_data = self.fetch_and_cache()
            if remote_data is not None:
                remote_data.store()
                self._data = remote_data
                return self._data

        raise EnumStringNotFound(
            value=self.slug,
            enum=self.__class__.__name__,
            possible_values=list(self.iter()),
        )

    def load_from_runtime(self) -> t.Optional[DataT]:
        raise NotImplementedError

    def fetch_and_cache(self) -> t.Optional[DataT]:
        raise NotImplementedError


class EnumGenerator(type):
    def __getattr__(cls, name: str):
        return cls(name)



"""
Enum helper base.
"""

import difflib
import typing as t

from composio.constants import LOCAL_CACHE_DIRECTORY
from composio.exceptions import ComposioSDKError
from composio.storage.base import LocalStorage


_runtime_actions: t.Dict[str, "ActionData"] = {}

EntityType = t.TypeVar("EntityType", bound=LocalStorage)

TAGS_CACHE = LOCAL_CACHE_DIRECTORY / "tags"
APPS_CACHE = LOCAL_CACHE_DIRECTORY / "apps"
ACTIONS_CACHE = LOCAL_CACHE_DIRECTORY / "actions"
TRIGGERS_CACHE = LOCAL_CACHE_DIRECTORY / "triggers"


class EnumStringNotFound(ComposioSDKError):
    """Raise when user provides invalid enum string."""

    def __init__(self, value: str, enum: str, possible_values: t.List[str]) -> None:
        error_message = f"Invalid value `{value}` for enum class `{enum}`"
        matches = difflib.get_close_matches(value, possible_values, n=1)
        if matches:
            (match,) = matches
            error_message += f". Did you mean {match!r}?"

        super().__init__(message=error_message)


class SentinalObject:
    """Sentinel object."""

    sentinel = None


class TagData(LocalStorage):
    """Local storage for `Tag` object."""

    app: str
    "App name for this tag."

    value: str
    "Tag string."


class AppData(LocalStorage):
    """Local storage for `App` object."""

    name: str
    "Name of the app."

    is_local: bool = False
    "The tool is local if set to `True`"


class ActionData(LocalStorage):
    """Local storage for `Action` object."""

    name: str
    "Action name."

    app: str
    "App name where the actions belongs to."

    tags: t.List[str]
    "Tag string for the action."

    no_auth: bool = False
    "If set `True` the action does not require authentication."

    is_local: bool = False
    "If set `True` the `app` is a local app."

    is_runtime: bool = False
    "Set `True` for actions registered at runtime."

    shell: bool = False
    "If set `True` the action will be executed using a shell."

    replaced_by: t.Optional[str] = None
    "If set, the action is deprecated and replaced by the given action."


class TriggerData(LocalStorage):
    """Local storage for `Trigger` object."""

    name: str
    "Name of the trigger."

    app: str
    "Name of the app where this trigger belongs to."


def add_runtime_action(name: str, data: ActionData) -> None:
    """Add action at runtime."""
    _runtime_actions[name] = data


# TODO: action_registry exists, how is _runtime_actions different from that?
def get_runtime_actions() -> t.List:
    """Add action at runtime."""
    return list(_runtime_actions)


DEPRECATED_MARKER = "<<DEPRECATED use "


def replacement_action_name(description: str, app_name: str) -> t.Optional[str]:
    """If the action is deprecated, get the replacement action name."""
    if description is not None and DEPRECATED_MARKER in description:
        _, newact = description.split(DEPRECATED_MARKER, maxsplit=1)
        return (app_name + "_" + newact.replace(">>", "")).upper()

    return None



import json
import os
import typing as t

from composio.client import Composio, enums
from composio.client.collections import ActionModel, AppModel, TriggerModel
from composio.client.enums.base import replacement_action_name
from composio.tools.local import load_local_tools
from composio.utils import get_enum_key
from composio.utils.logging import get_logger


EnumModels = t.Union[AppModel, ActionModel, TriggerModel]


logger = get_logger(__name__)

NO_CACHE_REFRESH = os.getenv("COMPOSIO_NO_CACHE_REFRESH", "false") == "true"


def filter_non_beta_items(items: t.Sequence[EnumModels]) -> t.List:
    filtered_items: t.List[EnumModels] = []
    for item in items:
        if not item.name.lower().endswith("beta"):
            filtered_items.append(item)

    seen: t.Set[str] = set()
    unique_items: t.List[EnumModels] = []
    for item in filtered_items:
        if item.name not in seen:
            unique_items.append(item)
            seen.add(item.name)

    return unique_items


def update_apps(client: Composio, beta: bool = False) -> t.List[AppModel]:
    """Update apps."""
    apps = sorted(
        client.apps.get(),
        key=lambda x: x.key,
    )
    if not beta:
        apps = filter_non_beta_items(apps)

    _update_apps(apps=apps)
    return apps


def update_actions(
    client: Composio, apps: t.List[AppModel], beta: bool = False
) -> None:
    """Update actions and tags."""
    actions = sorted(
        client.actions.get(allow_all=True),
        key=lambda x: f"{x.appName}_{x.name}",
    )
    if not beta:
        actions = filter_non_beta_items(actions)

    _update_tags(apps=apps, actions=actions)
    _update_actions(apps=apps, actions=actions)


def update_triggers(
    client: Composio, apps: t.List[AppModel], beta: bool = False
) -> None:
    """Update triggers."""
    triggers = sorted(
        client.triggers.get(),
        key=lambda x: f"{x.appKey}_{x.name}",
    )
    if not beta:
        triggers = filter_non_beta_items(triggers)

    _update_triggers(apps=apps, triggers=triggers)


def _update_apps(apps: t.List[AppModel]) -> None:
    """Create App enum class."""
    app_names = []
    enums.base.APPS_CACHE.mkdir(
        parents=True,
        exist_ok=True,
    )
    for app in apps:
        app_names.append(
            get_enum_key(
                name=app.key.lower().replace(" ", "_").replace("-", "_"),
            )
        )
        enums.base.AppData(
            name=app.name,
            path=enums.base.APPS_CACHE / app_names[-1],
            is_local=False,
        ).store()

    for tool in load_local_tools()["local"].values():
        if tool.enum in app_names:
            continue

        app_names.append(tool.enum)
        enums.base.AppData(
            name=tool.name,
            path=enums.base.APPS_CACHE / app_names[-1],
            is_local=True,
        ).store()


def _update_actions(apps: t.List[AppModel], actions: t.List[ActionModel]) -> None:
    """Get Action enum."""
    enums.base.ACTIONS_CACHE.mkdir(parents=True, exist_ok=True)
    deprecated = {}
    action_names = []
    for app in sorted(apps, key=lambda x: x.key):
        for action in actions:
            if action.appName != app.key:
                continue

            new_action_name = replacement_action_name(
                action.description or "", action.appName
            )
            if new_action_name is not None:
                replaced_by = deprecated[get_enum_key(name=action.name)] = (
                    new_action_name
                )
            else:
                action_names.append(get_enum_key(name=action.name))
                replaced_by = None

            # TODO: there is duplicate ActionData creation code in
            # `load_from_runtime` and `fetch_and_cache` in client/enums/action.py
            enums.base.ActionData(
                name=action.name,
                app=app.key,
                tags=action.tags,
                no_auth=app.no_auth,
                is_local=False,
                path=enums.base.ACTIONS_CACHE / get_enum_key(name=action.name),
                replaced_by=replaced_by,
            ).store()

    processed = []
    for tool in load_local_tools()["local"].values():
        if tool.name in processed:
            continue

        processed.append(tool.name)
        for actcls in tool.actions():
            action_names.append(actcls.enum)
            enums.base.ActionData(
                name=actcls.enum,
                app=tool.name,
                tags=actcls.tags(),
                no_auth=True,
                is_local=True,
                path=enums.base.ACTIONS_CACHE / action_names[-1],
                shell=False,
            ).store()


def _update_tags(apps: t.List[AppModel], actions: t.List[ActionModel]) -> None:
    """Create Tag enum class."""
    enums.base.TAGS_CACHE.mkdir(parents=True, exist_ok=True)
    tag_map: t.Dict[str, t.Set[str]] = {}
    for app in apps:
        app_name = app.key
        for action in [action for action in actions if action.appName == app_name]:
            if app_name not in tag_map:
                tag_map[app_name] = set()
            tag_map[app_name].update(action.tags or [])

    tag_names = ["DEFAULT"]
    for app_name in sorted(tag_map):
        for tag in sorted(tag_map[app_name]):
            tag_name = get_enum_key(name=f"{app_name}_{tag}")
            tag_names.append(tag_name)
            enums.base.TagData(
                app=app_name,
                value=tag,
                path=enums.base.TAGS_CACHE / tag_names[-1],
            ).store()

    enums.base.TagData(
        app="default",
        value="important",
        path=enums.base.TAGS_CACHE / "DEFAULT",
    )


def _update_triggers(
    apps: t.List[AppModel],
    triggers: t.List[TriggerModel],
) -> None:
    """Get Trigger enum."""
    trigger_names = []
    enums.base.TRIGGERS_CACHE.mkdir(exist_ok=True)
    for app in apps:
        for trigger in triggers:
            if trigger.appKey != app.key:
                continue

            trigger_names.append(get_enum_key(name=trigger.name).upper())
            enums.base.TriggerData(
                name=trigger.name,
                app=app.key,
                path=enums.base.TRIGGERS_CACHE / trigger_names[-1],
            ).store()


def check_cache_refresh(client: Composio) -> None:
    """
    Check if the actions have a 'replaced_by' field and refresh the cache if not.
    This is a workaround to invalidate local caches from older Composio versionos
    that do not have the 'replaced_by' field.

    Before this version, checking if an action is deprecated or not depended on the
    SDK version, and didn't come from the API. We need to start storing the data
    from the API and invalidate the cache if the data is not already stored.
    """
    if NO_CACHE_REFRESH:
        return

    if enums.base.ACTIONS_CACHE.exists():
        first_file = next(enums.base.ACTIONS_CACHE.iterdir(), None)
        if first_file is not None:
            first_action = json.loads(first_file.read_text())
            if "replaced_by" in first_action:
                logger.debug("Actions cache is up-to-date")
                return

    logger.info("Actions cache is outdated, refreshing cache...")
    apps = update_apps(client)
    update_actions(client, apps)
    update_triggers(client, apps)



import sqlite3
from pathlib import Path
from typing import Dict

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


class SqlQueryRequest(BaseModel):
    query: str = Field(
        ...,
        description="SQL query to be executed",
    )
    connection_string: str = Field(
        ...,
        description="Database connection string",
    )


class SqlQueryResponse(BaseModel):
    execution_details: dict = Field(..., description="Execution details")
    response_data: list = Field(..., description="Result after executing the query")


class SqlQuery(LocalAction[SqlQueryRequest, SqlQueryResponse]):
    """
    Executes a SQL Query and returns the results for both local SQLite and remote databases
    """

    _tags = ["sql", "sql_query"]

    def _is_sqlite_connection(self, connection_string: str) -> bool:
        """Determine if the connection string is for a SQLite database"""
        return (
            connection_string.endswith(".db")
            or connection_string.endswith(".sqlite")
            or connection_string.endswith(".sqlite3")
            or connection_string.startswith("sqlite:///")
        )

    def execute(self, request: SqlQueryRequest, metadata: Dict) -> SqlQueryResponse:
        """Execute SQL query for either SQLite or remote databases"""
        import sqlalchemy.exc  # pylint: disable=import-outside-toplevel

        try:
            if self._is_sqlite_connection(request.connection_string):
                return self._execute_sqlite(request)

            return self._execute_remote(request)
        except sqlite3.Error as e:
            raise ValueError(f"SQLite database error: {str(e)}") from e
        except sqlalchemy.exc.SQLAlchemyError as e:
            raise ValueError(f"Database connection error: {str(e)}") from e
        except Exception as e:
            raise ValueError(f"Unexpected error: {str(e)}") from e

    def _execute_sqlite(self, request: SqlQueryRequest) -> SqlQueryResponse:
        """Execute query for SQLite database"""
        db_path = request.connection_string.replace("sqlite:///", "")
        if not Path(db_path).exists():
            raise ValueError(f"Error: Database file '{db_path}' does not exist.")
        with sqlite3.connect(db_path) as connection:
            cursor = connection.cursor()
            cursor.execute(request.query)
            response_data = [list(row) for row in cursor.fetchall()]
            connection.commit()
        return SqlQueryResponse(
            execution_details={"executed": True, "type": "sqlite"},
            response_data=response_data,
        )

    def _execute_remote(self, request: SqlQueryRequest) -> SqlQueryResponse:
        """Execute query for remote databases"""
        import sqlalchemy  # pylint: disable=import-outside-toplevel

        engine = sqlalchemy.create_engine(
            request.connection_string,
            pool_size=5,
            max_overflow=10,
            pool_timeout=30,
            pool_recycle=3600,
            connect_args={"connect_timeout": 10},
        )
        with engine.connect() as connection:
            result = connection.execute(sqlalchemy.text(request.query), {})
            response_data = [list(row) for row in result.fetchall()]
        return SqlQueryResponse(
            execution_details={"executed": True, "type": "remote"},
            response_data=response_data,
        )



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class RemoveGuestFromTaskRequest(BaseModel):
    """Request schema for `RemoveGuestFromTask`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )
    guest_id: int = Field(
        ...,
        alias="guest_id",
        description="",
    )
    include_shared: t.Optional[bool] = Field(
        default=None,
        alias="include_shared",
        description=(
            "Exclude details of items shared with the guest by setting this parameter "
            "to `false`. By default this parameter is set to `true`. "
        ),
    )
    custom_task_ids: t.Optional[bool] = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: t.Optional[int] = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )


class RemoveGuestFromTaskResponse(BaseModel):
    """Response schema for `RemoveGuestFromTask`"""

    data: t.Dict[str, t.Any]


class RemoveGuestFromTask(OpenAPIAction):
    """
    Revoke a guest's access to a task.    ***Note:** This endpoint is only available
    to Workspaces on our [Enterprise Plan](https://clickup.com/pricing).*
    """

    _tags = ["Guests"]
    _display_name = "remove_guest_from_task"
    _request_schema = RemoveGuestFromTaskRequest
    _response_schema = RemoveGuestFromTaskResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/guest/{guest_id}"
    method = "delete"
    operation_id = "Guests_revokeAccessToTask"
    action_identifier = "/task/{task_id}/guest/{guest_id}_delete"

    path_params = {"task_id": "task_id", "guest_id": "guest_id"}
    query_params = {
        "include_shared": "include_shared",
        "custom_task_ids": "custom_task_ids",
        "team_id": "team_id",
    }
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from composio.client.enums.enum import Enum, EnumGenerator

from .base import TriggerData

_TRIGGER_CACHE: t.Dict[str, "Trigger"] = {}

class Trigger(Enum[TriggerData], metaclass=EnumGenerator):
    cache_folder = "triggers"
    cache = _TRIGGER_CACHE
    storage = TriggerData

    def load_from_runtime(self) -> t.Optional[TriggerData]: ...
    def fetch_and_cache(self) -> t.Optional[TriggerData]: ...
    @property
    def name(self) -> str: ...
    @property
    def app(self) -> str: ...
    ASANA_TASK_TRIGGER: "Trigger"
    GITHUB_COMMIT_EVENT: "Trigger"
    GITHUB_FOLLOWER_EVENT: "Trigger"
    GITHUB_ISSUE_ADDED_EVENT: "Trigger"
    GITHUB_LABEL_ADDED_EVENT: "Trigger"
    GITHUB_PULL_REQUEST_EVENT: "Trigger"
    GITHUB_STAR_ADDED_EVENT: "Trigger"
    GMAIL_NEW_GMAIL_MESSAGE: "Trigger"
    GOOGLEDRIVE_GOOGLE_DRIVE_CHANGES: "Trigger"
    HUBSPOT_CONTACT_CREATED_TRIGGER: "Trigger"
    HUBSPOT_DEAL_STAGE_UPDATED_TRIGGER: "Trigger"
    HUBSPOT_FORM_SUBMITTED: "Trigger"
    LINEAR_COMMENT_EVENT_TRIGGER: "Trigger"
    LINEAR_ISSUE_CREATED_TRIGGER: "Trigger"
    LINEAR_ISSUE_UPDATED_TRIGGER: "Trigger"
    MAILCHIMP_MAILCHIMP_CAMPAIGN_TRIGGER: "Trigger"
    MAILCHIMP_MAILCHIMP_PROFILE_UPDATE_TRIGGER: "Trigger"
    MAILCHIMP_MAILCHIMP_SUBSCRIBE_TRIGGER: "Trigger"
    MAILCHIMP_MAILCHIMP_UNSUBSCRIBE_TRIGGER: "Trigger"
    NOTION_PAGE_ADDED_TO_DATABASE: "Trigger"
    NOTION_PAGE_ADDED_TRIGGER: "Trigger"
    NOTION_PAGE_UPDATED_TRIGGER: "Trigger"
    ONE_DRIVE_ONE_DRIVE_DELETED_TRIGGER: "Trigger"
    ONE_DRIVE_ONE_DRIVE_FILE_CREATED_TRIGGER: "Trigger"
    ONE_DRIVE_ONE_DRIVE_FOLDER_CREATED_TRIGGER: "Trigger"
    ONE_DRIVE_ONE_DRIVE_UPDATED_TRIGGER: "Trigger"
    OUTLOOK_OUTLOOK_CONTACT_TRIGGER: "Trigger"
    OUTLOOK_OUTLOOK_EVENT_TRIGGER: "Trigger"
    OUTLOOK_OUTLOOK_MESSAGE_TRIGGER: "Trigger"
    OUTLOOK_OUTLOOK_SENT_MESSAGE_TRIGGER: "Trigger"
    PIPEDRIVE_PIPEDRIVE_NEW_DEAL_TRIGGER: "Trigger"
    PIPEDRIVE_PIPEDRIVE_NEW_NOTE_TRIGGER: "Trigger"
    PIPEDRIVE_PIPEDRIVE_NEW_ORGANIZATION_TRIGGER: "Trigger"
    SALESFORCE_NEW_CONTACT_TRIGGER: "Trigger"
    SALESFORCE_NEW_LEAD_TRIGGER: "Trigger"
    SLACKBOT_CHANNEL_CREATED: "Trigger"
    SLACKBOT_REACTION_ADDED: "Trigger"
    SLACKBOT_REACTION_REMOVED: "Trigger"
    SLACKBOT_RECEIVE_MESSAGE: "Trigger"
    SLACKBOT_RECEIVE_THREAD_REPLY: "Trigger"
    SLACK_CHANNEL_CREATED: "Trigger"
    SLACK_REACTION_ADDED: "Trigger"
    SLACK_REACTION_REMOVED: "Trigger"
    SLACK_RECEIVE_MESSAGE: "Trigger"
    SLACK_RECEIVE_THREAD_REPLY: "Trigger"
    SPOTIFY_NEW_DEVICE_TRIGGER: "Trigger"
    SPOTIFY_PLAYLIST_ITEM_TRIGGER: "Trigger"
    SPOTIFY_PLAYLIST_TRIGGER: "Trigger"
    STRIPE_STRIPE_CHARGE_FAILED_TRIGGER: "Trigger"
    STRIPE_STRIPE_CHECKOUT_SESSION_COMPLETED_TRIGGER: "Trigger"
    STRIPE_STRIPE_INVOICE_PAYMENT_SUCCEEDED_TRIGGER: "Trigger"
    STRIPE_STRIPE_PAYMENT_FAILED_TRIGGER: "Trigger"
    STRIPE_STRIPE_PRODUCT_CREATED_TRIGGER: "Trigger"
    STRIPE_STRIPE_SUBSCRIPTION_ADDED_TRIGGER: "Trigger"
    STRIPE_STRIPE_SUBSCRIPTION_DELETED_TRIGGER: "Trigger"
    TRELLO_TRELLO_ARCHIVED_CARD_TRIGGER: "Trigger"
    TRELLO_TRELLO_NEW_BOARD_TRIGGER: "Trigger"
    TRELLO_TRELLO_NEW_CARD_TRIGGER: "Trigger"
    TRELLO_TRELLO_UPDATED_CARD_TRIGGER: "Trigger"
    YOUTUBE_NEW_ACTIVITY_TRIGGER: "Trigger"
    YOUTUBE_NEW_ITEM_IN_PLAYLIST_TRIGGER: "Trigger"
    YOUTUBE_NEW_PLAYLIST_TRIGGER: "Trigger"
    YOUTUBE_NEW_SUBSCRIPTION_TRIGGER: "Trigger"
    ZENDESK_NEW_USER_TRIGGER: "Trigger"
    ZENDESK_NEW_ZENDESK_TICKET_TRIGGER: "Trigger"



import typing as t

from composio.client.enums.action import Action
from composio.client.enums.enum import Enum, EnumGenerator

from .base import AppData


_APP_CACHE: t.Dict[str, "App"] = {}


class App(Enum[AppData], metaclass=EnumGenerator):
    cache_folder = "apps"
    cache = _APP_CACHE
    storage = AppData

    def load_from_runtime(self) -> t.Optional[AppData]:
        # re-load local tools, and check if it's a runtime app
        from composio.tools.base.abs import (  # pylint: disable=import-outside-toplevel
            tool_registry,
        )
        from composio.tools.local import (  # pylint: disable=import-outside-toplevel
            load_local_tools,
        )

        load_local_tools()

        for gid, tools in tool_registry.items():
            if self.slug in tools:
                self._data = AppData(
                    name=tools[self.slug].name,
                    is_local=gid in ("runtime", "local"),
                    path=self.storage_path,
                )
                return self._data

        return None

    def fetch_and_cache(self) -> t.Optional[AppData]:
        from composio.client import Composio  # pylint: disable=import-outside-toplevel

        client = Composio.get_latest()
        response = client.http.get(url=str(client.apps.endpoint / self.slug)).json()
        # TOFIX: Return proper error code when of item is not found
        if "message" in response:
            return None

        return AppData(
            name=response["name"],
            path=self.storage_path,
            is_local=False,
        )

    @property
    def name(self) -> str:
        return self.load().name

    @property
    def is_local(self) -> bool:
        return self.load().is_local

    def get_actions(self, tags: t.Optional[t.List[str]] = None) -> t.Iterator[Action]:
        """
        Get actions for the given app filtered by the `tags`

        :param tags: List of tags to filter the actions
        :return: Iterator object which yields `Action`
        """
        tags = tags or []
        for action in Action.all():
            if not action.slug.startswith(f"{self.slug}_"):
                continue

            if len(tags) == 0 or any(tag in action.tags for tag in tags):
                yield action



import typing as t

from composio.client.enums.enum import Enum, EnumGenerator

from .base import TagData

_TAG_CACHE: t.Dict[str, "Tag"] = {}

class Tag(Enum[TagData], metaclass=EnumGenerator):
    cache_folder = "tags"
    cache = _TAG_CACHE
    storage = TagData

    def fetch_and_cache(self) -> t.Optional[TagData]: ...
    def load_from_runtime(self) -> t.Optional[TagData]: ...
    @property
    def value(self) -> str: ...
    @property
    def app(self) -> str: ...
    AFFINITY_AUTH: "Tag"
    AFFINITY_COMPANIES: "Tag"
    AFFINITY_IMPORTANT: "Tag"
    AFFINITY_LISTS: "Tag"
    AFFINITY_OPPORTUNITIES: "Tag"
    AFFINITY_PERSONS: "Tag"
    AGENCYZOOM_AUTHENTICATION: "Tag"
    AGENCYZOOM_BUSINESS_CLASSIFICATIONS: "Tag"
    AGENCYZOOM_CONFIGURATION_AND_SETTINGS: "Tag"
    AGENCYZOOM_CONTACT_MANAGEMENT: "Tag"
    AGENCYZOOM_CUSTOMER_MANAGEMENT: "Tag"
    AGENCYZOOM_DEPARTMENT_AND_GROUPS: "Tag"
    AGENCYZOOM_IMPORTANT: "Tag"
    AGENCYZOOM_LEAD_MANAGEMENT: "Tag"
    AGENCYZOOM_LIFE_AND_HEALTH_LEAD_MANAGEMENT: "Tag"
    AGENCYZOOM_OPPORTUNITY_MANAGEMENT: "Tag"
    AGENCYZOOM_PIPELINE: "Tag"
    AGENCYZOOM_SERVICE_CENTER: "Tag"
    AGENCYZOOM_TASK_MANAGEMENT: "Tag"
    AGENCYZOOM_THREAD: "Tag"
    AGENCYZOOM_USER_MANAGEMENT: "Tag"
    AGENCYZOOM_V4SSO_AUTHENTICATION: "Tag"
    AHREFS_BACKLINKS_PROFILE: "Tag"
    AHREFS_BATCH_ANALYSIS: "Tag"
    AHREFS_CRAWLER: "Tag"
    AHREFS_IMPORTANT: "Tag"
    AHREFS_KEYWORD_IDEAS: "Tag"
    AHREFS_LIMITS_AND_USAGE: "Tag"
    AHREFS_ORGANIC_SEARCH: "Tag"
    AHREFS_OUTGOING_LINKS: "Tag"
    AHREFS_OVERVIEW: "Tag"
    AHREFS_PAGES: "Tag"
    AHREFS_PAID_SEARCH: "Tag"
    AHREFS_PROJECTS: "Tag"
    AHREFS_SERP_OVERVIEW: "Tag"
    AIRTABLE_IMPORTANT: "Tag"
    APOLLO_ACCOUNT: "Tag"
    APOLLO_APOLLO: "Tag"
    APOLLO_CONTACT: "Tag"
    APOLLO_CREATE: "Tag"
    APOLLO_DEAL: "Tag"
    APOLLO_DEALS: "Tag"
    APOLLO_ENRICHMENT: "Tag"
    APOLLO_IMPORTANT: "Tag"
    APOLLO_ORGANIZATION: "Tag"
    APOLLO_PEOPLE: "Tag"
    APOLLO_SEARCH: "Tag"
    APOLLO_STAGES: "Tag"
    APOLLO_TASK: "Tag"
    APOLLO_UPDATE: "Tag"
    APOLLO_USERS: "Tag"
    ASANA_ALLOCATIONS: "Tag"
    ASANA_ATTACHMENTS: "Tag"
    ASANA_AUDIT_LOG_API: "Tag"
    ASANA_BATCH_API: "Tag"
    ASANA_CUSTOM_FIELDS: "Tag"
    ASANA_CUSTOM_FIELD_SETTINGS: "Tag"
    ASANA_EVENTS: "Tag"
    ASANA_GOALS: "Tag"
    ASANA_GOAL_RELATIONSHIPS: "Tag"
    ASANA_IMPORTANT: "Tag"
    ASANA_JOBS: "Tag"
    ASANA_MEMBERSHIPS: "Tag"
    ASANA_ORGANIZATION_EXPORTS: "Tag"
    ASANA_PORTFOLIOS: "Tag"
    ASANA_PORTFOLIO_MEMBERSHIPS: "Tag"
    ASANA_PROJECTS: "Tag"
    ASANA_PROJECT_BRIEFS: "Tag"
    ASANA_PROJECT_MEMBERSHIPS: "Tag"
    ASANA_PROJECT_STATUSES: "Tag"
    ASANA_PROJECT_TEMPLATES: "Tag"
    ASANA_RULES: "Tag"
    ASANA_SECTIONS: "Tag"
    ASANA_STATUS_UPDATES: "Tag"
    ASANA_STORIES: "Tag"
    ASANA_TAGS: "Tag"
    ASANA_TASKS: "Tag"
    ASANA_TASK_TEMPLATES: "Tag"
    ASANA_TEAMS: "Tag"
    ASANA_TEAM_MEMBERSHIPS: "Tag"
    ASANA_TIME_PERIODS: "Tag"
    ASANA_TIME_TRACKING_ENTRIES: "Tag"
    ASANA_TYPEAHEAD: "Tag"
    ASANA_USERS: "Tag"
    ASANA_USER_TASK_LISTS: "Tag"
    ASANA_WEBHOOKS: "Tag"
    ASANA_WORKSPACES: "Tag"
    ASANA_WORKSPACE_MEMBERSHIPS: "Tag"
    ATTIO_ATTRIBUTES: "Tag"
    ATTIO_COMMENTS: "Tag"
    ATTIO_ENTRIES: "Tag"
    ATTIO_IMPORTANT: "Tag"
    ATTIO_LISTS: "Tag"
    ATTIO_META: "Tag"
    ATTIO_NOTES: "Tag"
    ATTIO_OBJECTS: "Tag"
    ATTIO_RECORDS: "Tag"
    ATTIO_TASKS: "Tag"
    ATTIO_THREADS: "Tag"
    ATTIO_WEBHOOKS: "Tag"
    ATTIO_WORKSPACE_MEMBERS: "Tag"
    BAMBOOHR_ACCOUNT_INFORMATION: "Tag"
    BAMBOOHR_APPLICANT_TRACKING: "Tag"
    BAMBOOHR_BENEFITS: "Tag"
    BAMBOOHR_COMPANY_FILES: "Tag"
    BAMBOOHR_EMPLOYEES: "Tag"
    BAMBOOHR_EMPLOYEE_FILES: "Tag"
    BAMBOOHR_GOALS: "Tag"
    BAMBOOHR_HOURS: "Tag"
    BAMBOOHR_IMPORTANT: "Tag"
    BAMBOOHR_LAST_CHANGE_INFORMATION: "Tag"
    BAMBOOHR_LOGIN: "Tag"
    BAMBOOHR_PAYROLL: "Tag"
    BAMBOOHR_PHOTOS: "Tag"
    BAMBOOHR_REPORTS: "Tag"
    BAMBOOHR_TABULAR_DATA: "Tag"
    BAMBOOHR_TIME_OFF: "Tag"
    BAMBOOHR_TIME_TRACKING: "Tag"
    BAMBOOHR_TIME_TRACKING___PRIVATE_BETA: "Tag"
    BAMBOOHR_TRAINING: "Tag"
    BAMBOOHR_WEBHOOKS: "Tag"
    BITBUCKET_ADDON: "Tag"
    BITBUCKET_BRANCHING_MODEL: "Tag"
    BITBUCKET_BRANCH_RESTRICTIONS: "Tag"
    BITBUCKET_COMMITS: "Tag"
    BITBUCKET_COMMIT_STATUSES: "Tag"
    BITBUCKET_DEPLOYMENTS: "Tag"
    BITBUCKET_DOWNLOADS: "Tag"
    BITBUCKET_IMPORTANT: "Tag"
    BITBUCKET_ISSUE_TRACKER: "Tag"
    BITBUCKET_PIPELINES: "Tag"
    BITBUCKET_PROJECTS: "Tag"
    BITBUCKET_PROPERTIES: "Tag"
    BITBUCKET_PULLREQUESTS: "Tag"
    BITBUCKET_REFS: "Tag"
    BITBUCKET_REPORTS: "Tag"
    BITBUCKET_REPOSITORIES: "Tag"
    BITBUCKET_SEARCH: "Tag"
    BITBUCKET_SNIPPETS: "Tag"
    BITBUCKET_SOURCE: "Tag"
    BITBUCKET_SSH: "Tag"
    BITBUCKET_USERS: "Tag"
    BITBUCKET_WEBHOOKS: "Tag"
    BITBUCKET_WORKSPACES: "Tag"
    BLACKBOARD_ADAPTIVE_RELEASE: "Tag"
    BLACKBOARD_ANNOUNCEMENTS: "Tag"
    BLACKBOARD_ATTEMPT_RECEIPT: "Tag"
    BLACKBOARD_ATTENDANCE: "Tag"
    BLACKBOARD_CALENDAR: "Tag"
    BLACKBOARD_CONTENT: "Tag"
    BLACKBOARD_CONTENT_COLLECTION_RESOURCES: "Tag"
    BLACKBOARD_CONTENT_FILE_ATTACHMENTS: "Tag"
    BLACKBOARD_CONTENT_GROUP_ASSIGNMENTS: "Tag"
    BLACKBOARD_CONTENT_RESOURCES: "Tag"
    BLACKBOARD_CONTENT_REVIEW: "Tag"
    BLACKBOARD_COURSES: "Tag"
    BLACKBOARD_COURSE_ANNOUNCEMENTS: "Tag"
    BLACKBOARD_COURSE_ASSESSMENTS: "Tag"
    BLACKBOARD_COURSE_CATEGORIES: "Tag"
    BLACKBOARD_COURSE_GRADEBOOK_CATEGORIES: "Tag"
    BLACKBOARD_COURSE_GRADES: "Tag"
    BLACKBOARD_COURSE_GRADE_ATTEMPTS: "Tag"
    BLACKBOARD_COURSE_GRADE_NOTATIONS: "Tag"
    BLACKBOARD_COURSE_GRADING_PERIODS: "Tag"
    BLACKBOARD_COURSE_GROUPS: "Tag"
    BLACKBOARD_COURSE_GROUP_USERS: "Tag"
    BLACKBOARD_COURSE_MEMBERSHIPS: "Tag"
    BLACKBOARD_COURSE_MESSAGES: "Tag"
    BLACKBOARD_COURSE_TOC: "Tag"
    BLACKBOARD_DATA_SOURCES: "Tag"
    BLACKBOARD_DEPRECATED___COURSES: "Tag"
    BLACKBOARD_DEPRECATED___COURSE_GRADES: "Tag"
    BLACKBOARD_DEPRECATED___COURSE_GROUPS: "Tag"
    BLACKBOARD_DEPRECATED___COURSE_GROUP_USERS: "Tag"
    BLACKBOARD_DISCUSSIONS: "Tag"
    BLACKBOARD_GOALS: "Tag"
    BLACKBOARD_IMPORTANT: "Tag"
    BLACKBOARD_INSTITUTIONAL_HIERARCHY: "Tag"
    BLACKBOARD_INSTITUTIONAL_HIERARCHY_ADMINISTRATORS: "Tag"
    BLACKBOARD_LTI: "Tag"
    BLACKBOARD_OAUTH: "Tag"
    BLACKBOARD_PERFORMANCE_DASHBOARD: "Tag"
    BLACKBOARD_PROCTORING: "Tag"
    BLACKBOARD_PRONOUNS: "Tag"
    BLACKBOARD_ROLES: "Tag"
    BLACKBOARD_RUBRICS: "Tag"
    BLACKBOARD_RUBRIC_ASSOCIATIONS: "Tag"
    BLACKBOARD_RUBRIC_EVALUATIONS: "Tag"
    BLACKBOARD_SESSIONS: "Tag"
    BLACKBOARD_SIS_LOGS: "Tag"
    BLACKBOARD_SYSTEM: "Tag"
    BLACKBOARD_TERMS: "Tag"
    BLACKBOARD_UPLOADS: "Tag"
    BLACKBOARD_USERS: "Tag"
    BREVO_ACCOUNT: "Tag"
    BREVO_COMPANIES: "Tag"
    BREVO_CONTACTS: "Tag"
    BREVO_CONVERSATIONS: "Tag"
    BREVO_COUPONS: "Tag"
    BREVO_DEALS: "Tag"
    BREVO_DOMAINS: "Tag"
    BREVO_ECOMMERCE: "Tag"
    BREVO_EMAIL_CAMPAIGNS: "Tag"
    BREVO_EVENT: "Tag"
    BREVO_EXTERNAL_FEEDS: "Tag"
    BREVO_FILES: "Tag"
    BREVO_IMPORTANT: "Tag"
    BREVO_INBOUND_PARSING: "Tag"
    BREVO_MASTER_ACCOUNT: "Tag"
    BREVO_NOTES: "Tag"
    BREVO_PROCESS: "Tag"
    BREVO_RESELLER: "Tag"
    BREVO_SENDERS: "Tag"
    BREVO_SMS_CAMPAIGNS: "Tag"
    BREVO_TASKS: "Tag"
    BREVO_TRANSACTIONAL_EMAILS: "Tag"
    BREVO_TRANSACTIONAL_SMS: "Tag"
    BREVO_TRANSACTIONAL_WHATSAPP: "Tag"
    BREVO_USER: "Tag"
    BREVO_WEBHOOKS: "Tag"
    BREVO_WHATSAPP_CAMPAIGNS: "Tag"
    CALENDLY_ACTIVITY_LOG: "Tag"
    CALENDLY_AVAILABILITY: "Tag"
    CALENDLY_DATA_COMPLIANCE: "Tag"
    CALENDLY_EVENT_TYPES: "Tag"
    CALENDLY_GROUPS: "Tag"
    CALENDLY_IMPORTANT: "Tag"
    CALENDLY_ORGANIZATIONS: "Tag"
    CALENDLY_OUTGOING_COMMUNICATIONS: "Tag"
    CALENDLY_ROUTING_FORMS: "Tag"
    CALENDLY_SCHEDULED_EVENTS: "Tag"
    CALENDLY_SCHEDULING_LINKS: "Tag"
    CALENDLY_SHARES: "Tag"
    CALENDLY_USERS: "Tag"
    CALENDLY_WEBHOOKS: "Tag"
    CAL_BOOKINGS: "Tag"
    CAL_CALENDARS: "Tag"
    CAL_CONFERENCING: "Tag"
    CAL_DESTINATION_CALENDARS: "Tag"
    CAL_EVENT_TYPES: "Tag"
    CAL_EVENT_TYPES___WEBHOOKS: "Tag"
    CAL_IMPORTANT: "Tag"
    CAL_ME: "Tag"
    CAL_ORGS___ATTRIBUTES: "Tag"
    CAL_ORGS___ATTRIBUTES___OPTIONS: "Tag"
    CAL_ORGS___EVENT_TYPES: "Tag"
    CAL_ORGS___MEMBERSHIPS: "Tag"
    CAL_ORGS___SCHEDULES: "Tag"
    CAL_ORGS___TEAMS: "Tag"
    CAL_ORGS___TEAMS___MEMBERSHIPS: "Tag"
    CAL_ORGS___TEAMS___SCHEDULES: "Tag"
    CAL_ORGS___TEAMS___USERS___SCHEDULES: "Tag"
    CAL_ORGS___USERS: "Tag"
    CAL_ORGS___USERS___SCHEDULES: "Tag"
    CAL_ORGS___WEBHOOKS: "Tag"
    CAL_PLATFORM___CAL_PROVIDER: "Tag"
    CAL_PLATFORM___GOOGLE_CALENDAR: "Tag"
    CAL_PLATFORM___MANAGED_USERS: "Tag"
    CAL_PLATFORM___WEBHOOKS: "Tag"
    CAL_SCHEDULES: "Tag"
    CAL_SELECTED_CALENDARS: "Tag"
    CAL_SLOTS: "Tag"
    CAL_STRIPE: "Tag"
    CAL_TEAMS: "Tag"
    CAL_TEAMS___EVENT_TYPES: "Tag"
    CAL_TEAMS___MEMBERSHIPS: "Tag"
    CAL_TIMEZONES: "Tag"
    CAL_WEBHOOKS: "Tag"
    CANVAS_ACCOUNTS: "Tag"
    CANVAS_ANALYTICS: "Tag"
    CANVAS_APPOINTMENTGROUPS: "Tag"
    CANVAS_ASSIGNMENTS: "Tag"
    CANVAS_CALENDAR: "Tag"
    CANVAS_COMMUNICATIONCHANNELS: "Tag"
    CANVAS_CONTENTEXPORTS: "Tag"
    CANVAS_CONTENTSHARES: "Tag"
    CANVAS_CONVERSATIONS: "Tag"
    CANVAS_COURSES: "Tag"
    CANVAS_ENROLLMENTS: "Tag"
    CANVAS_ENROLLMENTTERMS: "Tag"
    CANVAS_FILES: "Tag"
    CANVAS_FOLDERS: "Tag"
    CANVAS_GRADEBOOK_HISTORY: "Tag"
    CANVAS_IMPORTANT: "Tag"
    CANVAS_NOTIFICATIONS: "Tag"
    CANVAS_PAGES: "Tag"
    CANVAS_QUIZZES: "Tag"
    CANVAS_REPORTS: "Tag"
    CANVAS_SUBMISSIONS: "Tag"
    CANVAS_USERS: "Tag"
    CANVA_APP: "Tag"
    CANVA_ASSET: "Tag"
    CANVA_AUTOFILL: "Tag"
    CANVA_BRAND_TEMPLATE: "Tag"
    CANVA_COMMENT: "Tag"
    CANVA_CONNECT: "Tag"
    CANVA_DESIGN: "Tag"
    CANVA_DESIGN_IMPORT: "Tag"
    CANVA_EXPORT: "Tag"
    CANVA_FOLDER: "Tag"
    CANVA_IMPORTANT: "Tag"
    CANVA_OAUTH: "Tag"
    CANVA_USER: "Tag"
    CLICKUP_ATTACHMENTS: "Tag"
    CLICKUP_AUTHORIZATION: "Tag"
    CLICKUP_COMMENTS: "Tag"
    CLICKUP_CUSTOM_FIELDS: "Tag"
    CLICKUP_CUSTOM_TASK_TYPES: "Tag"
    CLICKUP_FOLDERS: "Tag"
    CLICKUP_GOALS: "Tag"
    CLICKUP_GUESTS: "Tag"
    CLICKUP_IMPORTANT: "Tag"
    CLICKUP_LISTS: "Tag"
    CLICKUP_MEMBERS: "Tag"
    CLICKUP_ROLES: "Tag"
    CLICKUP_SHARED_HIERARCHY: "Tag"
    CLICKUP_SPACES: "Tag"
    CLICKUP_TAGS: "Tag"
    CLICKUP_TASKS: "Tag"
    CLICKUP_TASK_CHECKLISTS: "Tag"
    CLICKUP_TASK_RELATIONSHIPS: "Tag"
    CLICKUP_TASK_TEMPLATES: "Tag"
    CLICKUP_TEAMS___USER_GROUPS: "Tag"
    CLICKUP_TEAMS___WORKSPACES: "Tag"
    CLICKUP_TIME_TRACKING: "Tag"
    CLICKUP_TIME_TRACKING__LEGACY_: "Tag"
    CLICKUP_USERS: "Tag"
    CLICKUP_VIEWS: "Tag"
    CLICKUP_WEBHOOKS: "Tag"
    DISCORDBOT_IMPORTANT: "Tag"
    DISCORD_IMPORTANT: "Tag"
    DOCUSIGN_ACCOUNTBRANDS: "Tag"
    DOCUSIGN_ACCOUNTCONSUMERDISCLOSURES: "Tag"
    DOCUSIGN_ACCOUNTCUSTOMFIELDS: "Tag"
    DOCUSIGN_ACCOUNTPASSWORDRULES: "Tag"
    DOCUSIGN_ACCOUNTPERMISSIONPROFILES: "Tag"
    DOCUSIGN_ACCOUNTS: "Tag"
    DOCUSIGN_ACCOUNTSEALPROVIDERS: "Tag"
    DOCUSIGN_ACCOUNTSIGNATUREPROVIDERS: "Tag"
    DOCUSIGN_ACCOUNTSIGNATURES: "Tag"
    DOCUSIGN_ACCOUNTTABSETTINGS: "Tag"
    DOCUSIGN_ACCOUNTWATERMARKS: "Tag"
    DOCUSIGN_AUTHORIZATIONS: "Tag"
    DOCUSIGN_BCCEMAILARCHIVE: "Tag"
    DOCUSIGN_BILLINGPLANS: "Tag"
    DOCUSIGN_BULKSEND: "Tag"
    DOCUSIGN_CHUNKEDUPLOADS: "Tag"
    DOCUSIGN_CLOUDSTORAGE: "Tag"
    DOCUSIGN_CLOUDSTORAGEPROVIDERS: "Tag"
    DOCUSIGN_COMMENTS: "Tag"
    DOCUSIGN_CONNECTCONFIGURATIONS: "Tag"
    DOCUSIGN_CONNECTEVENTS: "Tag"
    DOCUSIGN_CONTACTS: "Tag"
    DOCUSIGN_CUSTOMTABS: "Tag"
    DOCUSIGN_DOCUMENTRESPONSIVEHTMLPREVIEW: "Tag"
    DOCUSIGN_ENOTECONFIGURATIONS: "Tag"
    DOCUSIGN_ENVELOPEATTACHMENTS: "Tag"
    DOCUSIGN_ENVELOPECONSUMERDISCLOSURES: "Tag"
    DOCUSIGN_ENVELOPECUSTOMFIELDS: "Tag"
    DOCUSIGN_ENVELOPEDOCUMENTFIELDS: "Tag"
    DOCUSIGN_ENVELOPEDOCUMENTHTMLDEFINITIONS: "Tag"
    DOCUSIGN_ENVELOPEDOCUMENTS: "Tag"
    DOCUSIGN_ENVELOPEDOCUMENTVISIBILITY: "Tag"
    DOCUSIGN_ENVELOPEEMAILSETTINGS: "Tag"
    DOCUSIGN_ENVELOPEFORMDATA: "Tag"
    DOCUSIGN_ENVELOPEHTMLDEFINITIONS: "Tag"
    DOCUSIGN_ENVELOPELOCKS: "Tag"
    DOCUSIGN_ENVELOPEPUBLISH: "Tag"
    DOCUSIGN_ENVELOPERECIPIENTS: "Tag"
    DOCUSIGN_ENVELOPES: "Tag"
    DOCUSIGN_ENVELOPETEMPLATES: "Tag"
    DOCUSIGN_ENVELOPETRANSFERRULES: "Tag"
    DOCUSIGN_ENVELOPEVIEWS: "Tag"
    DOCUSIGN_ENVELOPEWORKFLOWDEFINITION: "Tag"
    DOCUSIGN_FAVORITETEMPLATES: "Tag"
    DOCUSIGN_GROUPBRANDS: "Tag"
    DOCUSIGN_GROUPS: "Tag"
    DOCUSIGN_GROUPUSERS: "Tag"
    DOCUSIGN_IDENTITYVERIFICATIONS: "Tag"
    DOCUSIGN_IMPORTANT: "Tag"
    DOCUSIGN_INVOICES: "Tag"
    DOCUSIGN_NOTARY: "Tag"
    DOCUSIGN_NOTARYJOURNALS: "Tag"
    DOCUSIGN_NOTARYJURISDICTION: "Tag"
    DOCUSIGN_PAYMENTGATEWAYACCOUNTS: "Tag"
    DOCUSIGN_PAYMENTS: "Tag"
    DOCUSIGN_POWERFORMDATA: "Tag"
    DOCUSIGN_POWERFORMS: "Tag"
    DOCUSIGN_REQUESTLOGS: "Tag"
    DOCUSIGN_RESOURCES: "Tag"
    DOCUSIGN_RESPONSIVEHTMLPREVIEW: "Tag"
    DOCUSIGN_SERVICES: "Tag"
    DOCUSIGN_SIGNINGGROUPS: "Tag"
    DOCUSIGN_SIGNINGGROUPUSERS: "Tag"
    DOCUSIGN_TABSBLOB: "Tag"
    DOCUSIGN_TEMPLATECUSTOMFIELDS: "Tag"
    DOCUSIGN_TEMPLATEDOCUMENTFIELDS: "Tag"
    DOCUSIGN_TEMPLATEDOCUMENTHTMLDEFINITIONS: "Tag"
    DOCUSIGN_TEMPLATEDOCUMENTRESPONSIVEHTMLPREVIEW: "Tag"
    DOCUSIGN_TEMPLATEDOCUMENTS: "Tag"
    DOCUSIGN_TEMPLATEDOCUMENTVISIBILITY: "Tag"
    DOCUSIGN_TEMPLATEHTMLDEFINITIONS: "Tag"
    DOCUSIGN_TEMPLATELOCKS: "Tag"
    DOCUSIGN_TEMPLATERECIPIENTS: "Tag"
    DOCUSIGN_TEMPLATERESPONSIVEHTMLPREVIEW: "Tag"
    DOCUSIGN_TEMPLATES: "Tag"
    DOCUSIGN_TEMPLATEVIEWS: "Tag"
    DOCUSIGN_USERCUSTOMSETTINGS: "Tag"
    DOCUSIGN_USERPROFILES: "Tag"
    DOCUSIGN_USERS: "Tag"
    DOCUSIGN_USERSIGNATURES: "Tag"
    DOCUSIGN_WORKSPACEITEMS: "Tag"
    DOCUSIGN_WORKSPACES: "Tag"
    ELEVENLABS_AUDIO_NATIVE: "Tag"
    ELEVENLABS_DUBBING: "Tag"
    ELEVENLABS_IMPORTANT: "Tag"
    ELEVENLABS_MODELS: "Tag"
    ELEVENLABS_PROJECTS: "Tag"
    ELEVENLABS_PRONUNCIATION_DICTIONARY: "Tag"
    ELEVENLABS_SAMPLES: "Tag"
    ELEVENLABS_SPEECH_HISTORY: "Tag"
    ELEVENLABS_SPEECH_TO_SPEECH: "Tag"
    ELEVENLABS_TEXT_TO_SPEECH: "Tag"
    ELEVENLABS_USER: "Tag"
    ELEVENLABS_VOICES: "Tag"
    ELEVENLABS_VOICE_GENERATION: "Tag"
    ELEVENLABS_WORKSPACE: "Tag"
    FIGMA_ACTIVITY_LOGS: "Tag"
    FIGMA_COMMENTS: "Tag"
    FIGMA_COMMENT_REACTIONS: "Tag"
    FIGMA_COMPONENTS: "Tag"
    FIGMA_COMPONENT_SETS: "Tag"
    FIGMA_DEV_RESOURCES: "Tag"
    FIGMA_FILES: "Tag"
    FIGMA_IMPORTANT: "Tag"
    FIGMA_PAYMENTS: "Tag"
    FIGMA_PROJECTS: "Tag"
    FIGMA_STYLES: "Tag"
    FIGMA_USERS: "Tag"
    FIGMA_VARIABLES: "Tag"
    FIGMA_WEBHOOKS: "Tag"
    FIRECRAWL_IMPORTANT: "Tag"
    GITHUB_ACTIONS: "Tag"
    GITHUB_ACTIVITY: "Tag"
    GITHUB_APPS: "Tag"
    GITHUB_BILLING: "Tag"
    GITHUB_CHECKS: "Tag"
    GITHUB_CLASSROOM: "Tag"
    GITHUB_CODESPACES: "Tag"
    GITHUB_CODES_OF_CONDUCT: "Tag"
    GITHUB_CODE_SCANNING: "Tag"
    GITHUB_COPILOT: "Tag"
    GITHUB_DEPENDABOT: "Tag"
    GITHUB_DEPENDENCY_GRAPH: "Tag"
    GITHUB_EMOJIS: "Tag"
    GITHUB_GISTS: "Tag"
    GITHUB_GIT: "Tag"
    GITHUB_GITIGNORE: "Tag"
    GITHUB_IMPORTANT: "Tag"
    GITHUB_INTERACTIONS: "Tag"
    GITHUB_ISSUES: "Tag"
    GITHUB_LICENSES: "Tag"
    GITHUB_MARKDOWN: "Tag"
    GITHUB_META: "Tag"
    GITHUB_MIGRATIONS: "Tag"
    GITHUB_OIDC: "Tag"
    GITHUB_ORGS: "Tag"
    GITHUB_PACKAGES: "Tag"
    GITHUB_PROJECTS: "Tag"
    GITHUB_PULLS: "Tag"
    GITHUB_RATE_LIMIT: "Tag"
    GITHUB_REACTIONS: "Tag"
    GITHUB_REPOS: "Tag"
    GITHUB_SEARCH: "Tag"
    GITHUB_SECRET_SCANNING: "Tag"
    GITHUB_SECURITY_ADVISORIES: "Tag"
    GITHUB_TEAMS: "Tag"
    GITHUB_USERS: "Tag"
    HEYGEN_CREATE_VIDEO_API: "Tag"
    HEYGEN_DEFAULT: "Tag"
    HEYGEN_IMPORTANT: "Tag"
    HEYGEN_LISTS: "Tag"
    HEYGEN_PERSONALIZED_VIDEO: "Tag"
    HEYGEN_STREAMING_API: "Tag"
    HEYGEN_TALKING_PHOTO: "Tag"
    HEYGEN_TEMPLATE_API: "Tag"
    HEYGEN_USER: "Tag"
    HEYGEN_VIDEO_TRANSLATE_API: "Tag"
    HEYGEN_WEBHOOKS: "Tag"
    HUBSPOT_BASIC: "Tag"
    HUBSPOT_BATCH: "Tag"
    HUBSPOT_CORE: "Tag"
    HUBSPOT_EVENTS: "Tag"
    HUBSPOT_GDPR: "Tag"
    HUBSPOT_GROUPS: "Tag"
    HUBSPOT_IMPORTANT: "Tag"
    HUBSPOT_OWNERS: "Tag"
    HUBSPOT_PIPELINES: "Tag"
    HUBSPOT_PIPELINE_AUDITS: "Tag"
    HUBSPOT_PIPELINE_STAGES: "Tag"
    HUBSPOT_PIPELINE_STAGE_AUDITS: "Tag"
    HUBSPOT_PUBLIC_IMPORTS: "Tag"
    HUBSPOT_PUBLIC_OBJECT: "Tag"
    HUBSPOT_PUBLIC_OBJECT_SCHEMAS: "Tag"
    HUBSPOT_RECORDING_SETTINGS: "Tag"
    HUBSPOT_SEARCH: "Tag"
    HUBSPOT_SETTINGS: "Tag"
    HUBSPOT_TEMPLATES: "Tag"
    HUBSPOT_TOKENS: "Tag"
    HUBSPOT_TYPES: "Tag"
    INTERCOM_ADMINS: "Tag"
    INTERCOM_ARTICLES: "Tag"
    INTERCOM_COMPANIES: "Tag"
    INTERCOM_CONTACTS: "Tag"
    INTERCOM_HELP_CENTER: "Tag"
    INTERCOM_IMPORTANT: "Tag"
    INTERCOM_NOTES: "Tag"
    INTERCOM_SEGMENTS: "Tag"
    INTERCOM_SUBSCRIPTION_TYPES: "Tag"
    INTERCOM_TAGS: "Tag"
    JIRA_ANNOUNCEMENT_BANNER: "Tag"
    JIRA_APPLICATION_ROLES: "Tag"
    JIRA_APP_DATA_POLICIES__EAP_: "Tag"
    JIRA_APP_MIGRATION: "Tag"
    JIRA_APP_PROPERTIES: "Tag"
    JIRA_AUDIT_RECORDS: "Tag"
    JIRA_AVATARS: "Tag"
    JIRA_CLASSIFICATION_LEVELS: "Tag"
    JIRA_DASHBOARDS: "Tag"
    JIRA_DYNAMIC_MODULES: "Tag"
    JIRA_FILTERS: "Tag"
    JIRA_FILTER_SHARING: "Tag"
    JIRA_GROUPS: "Tag"
    JIRA_GROUP_AND_USER_PICKER: "Tag"
    JIRA_IMPORTANT: "Tag"
    JIRA_ISSUES: "Tag"
    JIRA_ISSUE_ATTACHMENTS: "Tag"
    JIRA_ISSUE_COMMENTS: "Tag"
    JIRA_ISSUE_COMMENT_PROPERTIES: "Tag"
    JIRA_ISSUE_CUSTOM_FIELD_CONFIGURATION__APPS_: "Tag"
    JIRA_ISSUE_CUSTOM_FIELD_CONTEXTS: "Tag"
    JIRA_ISSUE_CUSTOM_FIELD_OPTIONS: "Tag"
    JIRA_ISSUE_CUSTOM_FIELD_OPTIONS__APPS_: "Tag"
    JIRA_ISSUE_CUSTOM_FIELD_VALUES__APPS_: "Tag"
    JIRA_ISSUE_FIELDS: "Tag"
    JIRA_ISSUE_FIELD_CONFIGURATIONS: "Tag"
    JIRA_ISSUE_LINKS: "Tag"
    JIRA_ISSUE_LINK_TYPES: "Tag"
    JIRA_ISSUE_NAVIGATOR_SETTINGS: "Tag"
    JIRA_ISSUE_NOTIFICATION_SCHEMES: "Tag"
    JIRA_ISSUE_PRIORITIES: "Tag"
    JIRA_ISSUE_PROPERTIES: "Tag"
    JIRA_ISSUE_REMOTE_LINKS: "Tag"
    JIRA_ISSUE_RESOLUTIONS: "Tag"
    JIRA_ISSUE_SEARCH: "Tag"
    JIRA_ISSUE_SECURITY_LEVEL: "Tag"
    JIRA_ISSUE_SECURITY_SCHEMES: "Tag"
    JIRA_ISSUE_TYPES: "Tag"
    JIRA_ISSUE_TYPE_PROPERTIES: "Tag"
    JIRA_ISSUE_TYPE_SCHEMES: "Tag"
    JIRA_ISSUE_TYPE_SCREEN_SCHEMES: "Tag"
    JIRA_ISSUE_VOTES: "Tag"
    JIRA_ISSUE_WATCHERS: "Tag"
    JIRA_ISSUE_WORKLOGS: "Tag"
    JIRA_ISSUE_WORKLOG_PROPERTIES: "Tag"
    JIRA_JIRA_EXPRESSIONS: "Tag"
    JIRA_JIRA_SETTINGS: "Tag"
    JIRA_JQL: "Tag"
    JIRA_JQL_FUNCTIONS__APPS_: "Tag"
    JIRA_LABELS: "Tag"
    JIRA_LICENSE_METRICS: "Tag"
    JIRA_MYSELF: "Tag"
    JIRA_PERMISSIONS: "Tag"
    JIRA_PERMISSION_SCHEMES: "Tag"
    JIRA_PROJECTS: "Tag"
    JIRA_PROJECT_AVATARS: "Tag"
    JIRA_PROJECT_CATEGORIES: "Tag"
    JIRA_PROJECT_CLASSIFICATION_LEVELS: "Tag"
    JIRA_PROJECT_COMPONENTS: "Tag"
    JIRA_PROJECT_EMAIL: "Tag"
    JIRA_PROJECT_FEATURES: "Tag"
    JIRA_PROJECT_KEY_AND_NAME_VALIDATION: "Tag"
    JIRA_PROJECT_PERMISSION_SCHEMES: "Tag"
    JIRA_PROJECT_PROPERTIES: "Tag"
    JIRA_PROJECT_ROLES: "Tag"
    JIRA_PROJECT_ROLE_ACTORS: "Tag"
    JIRA_PROJECT_TYPES: "Tag"
    JIRA_PROJECT_VERSIONS: "Tag"
    JIRA_SCREENS: "Tag"
    JIRA_SCREEN_SCHEMES: "Tag"
    JIRA_SCREEN_TABS: "Tag"
    JIRA_SCREEN_TAB_FIELDS: "Tag"
    JIRA_SERVER_INFO: "Tag"
    JIRA_SERVICE_REGISTRY: "Tag"
    JIRA_STATUS: "Tag"
    JIRA_TASKS: "Tag"
    JIRA_TIME_TRACKING: "Tag"
    JIRA_UI_MODIFICATIONS__APPS_: "Tag"
    JIRA_USERS: "Tag"
    JIRA_USER_PROPERTIES: "Tag"
    JIRA_USER_SEARCH: "Tag"
    JIRA_WEBHOOKS: "Tag"
    JIRA_WORKFLOWS: "Tag"
    JIRA_WORKFLOW_SCHEMES: "Tag"
    JIRA_WORKFLOW_SCHEME_DRAFTS: "Tag"
    JIRA_WORKFLOW_SCHEME_PROJECT_ASSOCIATIONS: "Tag"
    JIRA_WORKFLOW_STATUSES: "Tag"
    JIRA_WORKFLOW_STATUS_CATEGORIES: "Tag"
    JIRA_WORKFLOW_TRANSITION_PROPERTIES: "Tag"
    JIRA_WORKFLOW_TRANSITION_RULES: "Tag"
    KLAVIYO_ACCOUNTS: "Tag"
    KLAVIYO_CAMPAIGNS: "Tag"
    KLAVIYO_CATALOGS: "Tag"
    KLAVIYO_CLIENT: "Tag"
    KLAVIYO_COUPONS: "Tag"
    KLAVIYO_DATA_PRIVACY: "Tag"
    KLAVIYO_EVENTS: "Tag"
    KLAVIYO_FLOWS: "Tag"
    KLAVIYO_FORMS: "Tag"
    KLAVIYO_IMAGES: "Tag"
    KLAVIYO_IMPORTANT: "Tag"
    KLAVIYO_LISTS: "Tag"
    KLAVIYO_METRICS: "Tag"
    KLAVIYO_PROFILES: "Tag"
    KLAVIYO_REPORTING: "Tag"
    KLAVIYO_SEGMENTS: "Tag"
    KLAVIYO_TAGS: "Tag"
    KLAVIYO_TEMPLATES: "Tag"
    KLAVIYO_WEBHOOKS: "Tag"
    LISTENNOTES_DIRECTORY_API: "Tag"
    LISTENNOTES_IMPORTANT: "Tag"
    LISTENNOTES_INSIGHTS_API: "Tag"
    LISTENNOTES_PLAYLIST_API: "Tag"
    LISTENNOTES_PODCASTER_API: "Tag"
    LISTENNOTES_SEARCH_API: "Tag"
    LMNT_ACCOUNT: "Tag"
    LMNT_SPEECH: "Tag"
    LMNT_VOICES: "Tag"
    MAILCHIMP_ACCOUNTEXPORT: "Tag"
    MAILCHIMP_ACCOUNTEXPORTS: "Tag"
    MAILCHIMP_ACTIVITYFEED: "Tag"
    MAILCHIMP_AUTHORIZEDAPPS: "Tag"
    MAILCHIMP_AUTOMATIONS: "Tag"
    MAILCHIMP_BATCHES: "Tag"
    MAILCHIMP_BATCHWEBHOOKS: "Tag"
    MAILCHIMP_CAMPAIGNFOLDERS: "Tag"
    MAILCHIMP_CAMPAIGNS: "Tag"
    MAILCHIMP_CONNECTEDSITES: "Tag"
    MAILCHIMP_CONVERSATIONS: "Tag"
    MAILCHIMP_CUSTOMERJOURNEYS: "Tag"
    MAILCHIMP_ECOMMERCE: "Tag"
    MAILCHIMP_FACEBOOKADS: "Tag"
    MAILCHIMP_FILEMANAGER: "Tag"
    MAILCHIMP_IMPORTANT: "Tag"
    MAILCHIMP_LANDINGPAGES: "Tag"
    MAILCHIMP_LISTS: "Tag"
    MAILCHIMP_PING: "Tag"
    MAILCHIMP_REPORTING: "Tag"
    MAILCHIMP_REPORTS: "Tag"
    MAILCHIMP_ROOT: "Tag"
    MAILCHIMP_SEARCHCAMPAIGNS: "Tag"
    MAILCHIMP_SEARCHMEMBERS: "Tag"
    MAILCHIMP_SURVEYS: "Tag"
    MAILCHIMP_TEMPLATEFOLDERS: "Tag"
    MAILCHIMP_TEMPLATES: "Tag"
    MAILCHIMP_VERIFIEDDOMAINS: "Tag"
    MEM0_AGENTS: "Tag"
    MEM0_APPS: "Tag"
    MEM0_ENTITIES: "Tag"
    MEM0_EVENTS: "Tag"
    MEM0_IMPORTANT: "Tag"
    MEM0_MEMORIES: "Tag"
    MEM0_ORGANIZATIONS: "Tag"
    MEM0_PROJECT: "Tag"
    MEM0_PROJECTS: "Tag"
    MEM0_RUNS: "Tag"
    MEM0_STATS: "Tag"
    MEM0_USERS: "Tag"
    NEON_API_KEY: "Tag"
    NEON_BRANCH: "Tag"
    NEON_CONSUMPTION: "Tag"
    NEON_ENDPOINT: "Tag"
    NEON_IMPORTANT: "Tag"
    NEON_OPERATION: "Tag"
    NEON_ORGANIZATIONS: "Tag"
    NEON_PROJECT: "Tag"
    NEON_REGION: "Tag"
    NEON_USERS: "Tag"
    NOTION_IMPORTANT: "Tag"
    OUTLOOK_EMAIL: "Tag"
    OUTLOOK_IMPORTANT: "Tag"
    OUTLOOK_SEARCH: "Tag"
    PAGERDUTY_ABILITIES: "Tag"
    PAGERDUTY_ADD_ONS: "Tag"
    PAGERDUTY_ALERT_GROUPING_SETTINGS: "Tag"
    PAGERDUTY_ANALYTICS: "Tag"
    PAGERDUTY_AUDIT: "Tag"
    PAGERDUTY_AUTOMATION_ACTIONS: "Tag"
    PAGERDUTY_BUSINESS_SERVICES: "Tag"
    PAGERDUTY_CHANGE_EVENTS: "Tag"
    PAGERDUTY_ESCALATION_POLICIES: "Tag"
    PAGERDUTY_EVENT_ORCHESTRATIONS: "Tag"
    PAGERDUTY_EXTENSIONS: "Tag"
    PAGERDUTY_EXTENSION_SCHEMAS: "Tag"
    PAGERDUTY_IMPORTANT: "Tag"
    PAGERDUTY_INCIDENTS: "Tag"
    PAGERDUTY_INCIDENT_CUSTOM_FIELDS: "Tag"
    PAGERDUTY_INCIDENT_TYPES: "Tag"
    PAGERDUTY_INCIDENT_WORKFLOWS: "Tag"
    PAGERDUTY_LICENSES: "Tag"
    PAGERDUTY_LOG_ENTRIES: "Tag"
    PAGERDUTY_MAINTENANCE_WINDOWS: "Tag"
    PAGERDUTY_NOTIFICATIONS: "Tag"
    PAGERDUTY_OAUTH_DELEGATIONS: "Tag"
    PAGERDUTY_ON_CALLS: "Tag"
    PAGERDUTY_PAUSED_INCIDENT_REPORTS: "Tag"
    PAGERDUTY_PRIORITIES: "Tag"
    PAGERDUTY_RESPONSE_PLAYS: "Tag"
    PAGERDUTY_RULESETS: "Tag"
    PAGERDUTY_SCHEDULES: "Tag"
    PAGERDUTY_SERVICES: "Tag"
    PAGERDUTY_SERVICE_DEPENDENCIES: "Tag"
    PAGERDUTY_STANDARDS: "Tag"
    PAGERDUTY_STATUS_DASHBOARDS: "Tag"
    PAGERDUTY_STATUS_PAGES: "Tag"
    PAGERDUTY_TAGS: "Tag"
    PAGERDUTY_TEAMS: "Tag"
    PAGERDUTY_TEMPLATES: "Tag"
    PAGERDUTY_USERS: "Tag"
    PAGERDUTY_VENDORS: "Tag"
    PAGERDUTY_WEBHOOKS: "Tag"
    PAGERDUTY_WORKFLOW_INTEGRATIONS: "Tag"
    PEOPLEDATALABS_IMPORTANT: "Tag"
    PIPEDRIVE_ACTIVITIES: "Tag"
    PIPEDRIVE_ACTIVITYFIELDS: "Tag"
    PIPEDRIVE_ACTIVITYTYPES: "Tag"
    PIPEDRIVE_BILLING: "Tag"
    PIPEDRIVE_CALLLOGS: "Tag"
    PIPEDRIVE_CHANNELS: "Tag"
    PIPEDRIVE_CURRENCIES: "Tag"
    PIPEDRIVE_DEALFIELDS: "Tag"
    PIPEDRIVE_DEALS: "Tag"
    PIPEDRIVE_FILES: "Tag"
    PIPEDRIVE_FILTERS: "Tag"
    PIPEDRIVE_GOALS: "Tag"
    PIPEDRIVE_IMPORTANT: "Tag"
    PIPEDRIVE_ITEMSEARCH: "Tag"
    PIPEDRIVE_LEADLABELS: "Tag"
    PIPEDRIVE_LEADS: "Tag"
    PIPEDRIVE_LEADSOURCES: "Tag"
    PIPEDRIVE_LEGACYTEAMS: "Tag"
    PIPEDRIVE_MAILBOX: "Tag"
    PIPEDRIVE_MEETINGS: "Tag"
    PIPEDRIVE_NOTEFIELDS: "Tag"
    PIPEDRIVE_NOTES: "Tag"
    PIPEDRIVE_OAUTH: "Tag"
    PIPEDRIVE_ORGANIZATIONFIELDS: "Tag"
    PIPEDRIVE_ORGANIZATIONRELATIONSHIPS: "Tag"
    PIPEDRIVE_ORGANIZATIONS: "Tag"
    PIPEDRIVE_PERMISSIONSETS: "Tag"
    PIPEDRIVE_PERSONFIELDS: "Tag"
    PIPEDRIVE_PERSONS: "Tag"
    PIPEDRIVE_PIPELINES: "Tag"
    PIPEDRIVE_PRODUCTFIELDS: "Tag"
    PIPEDRIVE_PRODUCTS: "Tag"
    PIPEDRIVE_PROJECTS: "Tag"
    PIPEDRIVE_PROJECTTEMPLATES: "Tag"
    PIPEDRIVE_RECENTS: "Tag"
    PIPEDRIVE_ROLES: "Tag"
    PIPEDRIVE_STAGES: "Tag"
    PIPEDRIVE_SUBSCRIPTIONS: "Tag"
    PIPEDRIVE_TASKS: "Tag"
    PIPEDRIVE_USERCONNECTIONS: "Tag"
    PIPEDRIVE_USERS: "Tag"
    PIPEDRIVE_USERSETTINGS: "Tag"
    PIPEDRIVE_WEBHOOKS: "Tag"
    POSTHOG_ACTIONS: "Tag"
    POSTHOG_ACTIVITY_LOG: "Tag"
    POSTHOG_ANNOTATIONS: "Tag"
    POSTHOG_APP_METRICS: "Tag"
    POSTHOG_BATCH_EXPORTS: "Tag"
    POSTHOG_COHORTS: "Tag"
    POSTHOG_DASHBOARDS: "Tag"
    POSTHOG_DASHBOARD_TEMPLATES: "Tag"
    POSTHOG_DOMAINS: "Tag"
    POSTHOG_EARLY_ACCESS_FEATURE: "Tag"
    POSTHOG_EVENTS: "Tag"
    POSTHOG_EVENT_DEFINITIONS: "Tag"
    POSTHOG_EXPERIMENTS: "Tag"
    POSTHOG_EXPLICIT_MEMBERS: "Tag"
    POSTHOG_EXPORTS: "Tag"
    POSTHOG_FEATURE_FLAGS: "Tag"
    POSTHOG_FUNNEL: "Tag"
    POSTHOG_GROUPS: "Tag"
    POSTHOG_GROUPS_TYPES: "Tag"
    POSTHOG_IMPORTANT: "Tag"
    POSTHOG_INSIGHTS: "Tag"
    POSTHOG_INVITES: "Tag"
    POSTHOG_MEMBERS: "Tag"
    POSTHOG_NOTEBOOKS: "Tag"
    POSTHOG_ORGANIZATIONS: "Tag"
    POSTHOG_PERSONS: "Tag"
    POSTHOG_PIPELINE_DESTINATIONS: "Tag"
    POSTHOG_PIPELINE_DESTINATION_CONFIGS: "Tag"
    POSTHOG_PIPELINE_FRONTEND_APPS: "Tag"
    POSTHOG_PIPELINE_FRONTEND_APPS_CONFIGS: "Tag"
    POSTHOG_PIPELINE_IMPORT_APPS: "Tag"
    POSTHOG_PIPELINE_IMPORT_APPS_CONFIGS: "Tag"
    POSTHOG_PIPELINE_TRANSFORMATIONS: "Tag"
    POSTHOG_PIPELINE_TRANSFORMATION_CONFIGS: "Tag"
    POSTHOG_PLUGINS: "Tag"
    POSTHOG_PLUGIN_CONFIGS: "Tag"
    POSTHOG_PROJECTS: "Tag"
    POSTHOG_PROPERTY_DEFINITIONS: "Tag"
    POSTHOG_PROXY_RECORDS: "Tag"
    POSTHOG_QUERY: "Tag"
    POSTHOG_ROLES: "Tag"
    POSTHOG_SESSIONS: "Tag"
    POSTHOG_SESSION_RECORDINGS: "Tag"
    POSTHOG_SESSION_RECORDING_PLAYLISTS: "Tag"
    POSTHOG_SUBSCRIPTIONS: "Tag"
    POSTHOG_SURVEYS: "Tag"
    POSTHOG_TREND: "Tag"
    POSTHOG_USERS: "Tag"
    SALESFORCE_ACCOUNT: "Tag"
    SALESFORCE_CAMPAIGN: "Tag"
    SALESFORCE_CONTACT: "Tag"
    SALESFORCE_IMPORTANT: "Tag"
    SALESFORCE_LEAD: "Tag"
    SALESFORCE_NOTE: "Tag"
    SALESFORCE_OPPORTUNITY: "Tag"
    SENDGRID_ACCOUNT: "Tag"
    SENDGRID_ACCOUNT_STATE: "Tag"
    SENDGRID_ALERTS: "Tag"
    SENDGRID_API_KEYS: "Tag"
    SENDGRID_BLOCKS: "Tag"
    SENDGRID_BOUNCES: "Tag"
    SENDGRID_BULK_EMAIL_ADDRESS_VALIDATION: "Tag"
    SENDGRID_CAMPAIGNS_API: "Tag"
    SENDGRID_CATEGORIES: "Tag"
    SENDGRID_CONTACTS: "Tag"
    SENDGRID_CUSTOM_FIELDS: "Tag"
    SENDGRID_DESIGNS: "Tag"
    SENDGRID_DOMAIN_AUTHENTICATION: "Tag"
    SENDGRID_EMAIL_ACTIVITY: "Tag"
    SENDGRID_EMAIL_ADDRESS_VALIDATION: "Tag"
    SENDGRID_ENFORCED_TLS: "Tag"
    SENDGRID_ENGAGEMENT_QUALITY: "Tag"
    SENDGRID_EVENT_WEBHOOK: "Tag"
    SENDGRID_EXTERNAL_INTEGRATION_ENDPOINTS: "Tag"
    SENDGRID_GLOBAL_SUPPRESSIONS: "Tag"
    SENDGRID_IMPORTANT: "Tag"
    SENDGRID_INVALID_EMAILS: "Tag"
    SENDGRID_IP_ACCESS_MANAGEMENT: "Tag"
    SENDGRID_IP_ADDRESSES: "Tag"
    SENDGRID_IP_ADDRESS_MANAGEMENT: "Tag"
    SENDGRID_IP_POOLS: "Tag"
    SENDGRID_IP_WARMUP: "Tag"
    SENDGRID_LINK_BRANDING: "Tag"
    SENDGRID_LISTS: "Tag"
    SENDGRID_MAIL_BATCH: "Tag"
    SENDGRID_MAIL_SEND: "Tag"
    SENDGRID_MAIL_SETTINGS: "Tag"
    SENDGRID_OFFERING: "Tag"
    SENDGRID_PARSE_WEBHOOK: "Tag"
    SENDGRID_PARTNER_SETTINGS: "Tag"
    SENDGRID_POINT_DELETE_SYSTEM: "Tag"
    SENDGRID_RECIPIENTS: "Tag"
    SENDGRID_REVERSE_DNS: "Tag"
    SENDGRID_SCHEDULED_SENDS: "Tag"
    SENDGRID_SCOPES: "Tag"
    SENDGRID_SEGMENTING_CONTACTS: "Tag"
    SENDGRID_SEGMENTING_CONTACTS_V2: "Tag"
    SENDGRID_SEGMENTS: "Tag"
    SENDGRID_SENDERS: "Tag"
    SENDGRID_SENDER_IDENTITIES: "Tag"
    SENDGRID_SENDER_VERIFICATION: "Tag"
    SENDGRID_SEND_TEST_EMAIL: "Tag"
    SENDGRID_SINGLE_SENDS: "Tag"
    SENDGRID_SPAM_REPORTS: "Tag"
    SENDGRID_SSO_CERTIFICATES: "Tag"
    SENDGRID_SSO_SETTINGS: "Tag"
    SENDGRID_SSO_TEAMMATES: "Tag"
    SENDGRID_STATS: "Tag"
    SENDGRID_SUBUSERS: "Tag"
    SENDGRID_SUBUSER_STATISTICS: "Tag"
    SENDGRID_SUBUSER_WEBSITE_ACCESS: "Tag"
    SENDGRID_SUPPRESSIONS: "Tag"
    SENDGRID_TEAMMATES: "Tag"
    SENDGRID_TEMPLATES: "Tag"
    SENDGRID_TEMPLATES_VERSIONS: "Tag"
    SENDGRID_TRACKING: "Tag"
    SENDGRID_UNSUBSCRIBE_GROUPS: "Tag"
    SENDGRID_USERS_API: "Tag"
    SHOPIFY_IMPORTANT: "Tag"
    SHOPIFY_PRODUCT_IMAGE: "Tag"
    SLACKBOT_ADMIN: "Tag"
    SLACKBOT_ADMIN_APPS: "Tag"
    SLACKBOT_ADMIN_APPS_APPROVED: "Tag"
    SLACKBOT_ADMIN_APPS_REQUESTS: "Tag"
    SLACKBOT_ADMIN_APPS_RESTRICTED: "Tag"
    SLACKBOT_ADMIN_CONVERSATIONS: "Tag"
    SLACKBOT_ADMIN_CONVERSATIONS_EKM: "Tag"
    SLACKBOT_ADMIN_CONVERSATIONS_RESTRICTACCESS: "Tag"
    SLACKBOT_ADMIN_EMOJI: "Tag"
    SLACKBOT_ADMIN_INVITEREQUESTS: "Tag"
    SLACKBOT_ADMIN_INVITEREQUESTS_APPROVED: "Tag"
    SLACKBOT_ADMIN_INVITEREQUESTS_DENIED: "Tag"
    SLACKBOT_ADMIN_TEAMS: "Tag"
    SLACKBOT_ADMIN_TEAMS_ADMINS: "Tag"
    SLACKBOT_ADMIN_TEAMS_OWNERS: "Tag"
    SLACKBOT_ADMIN_TEAMS_SETTINGS: "Tag"
    SLACKBOT_ADMIN_USERGROUPS: "Tag"
    SLACKBOT_ADMIN_USERS: "Tag"
    SLACKBOT_ADMIN_USERS_SESSION: "Tag"
    SLACKBOT_API: "Tag"
    SLACKBOT_APPS: "Tag"
    SLACKBOT_APPS_EVENT_AUTHORIZATIONS: "Tag"
    SLACKBOT_APPS_PERMISSIONS: "Tag"
    SLACKBOT_APPS_PERMISSIONS_RESOURCES: "Tag"
    SLACKBOT_APPS_PERMISSIONS_SCOPES: "Tag"
    SLACKBOT_APPS_PERMISSIONS_USERS: "Tag"
    SLACKBOT_AUTH: "Tag"
    SLACKBOT_BOTS: "Tag"
    SLACKBOT_CALLS: "Tag"
    SLACKBOT_CALLS_PARTICIPANTS: "Tag"
    SLACKBOT_CHAT: "Tag"
    SLACKBOT_CHAT_SCHEDULEDMESSAGES: "Tag"
    SLACKBOT_CONVERSATIONS: "Tag"
    SLACKBOT_DIALOG: "Tag"
    SLACKBOT_DND: "Tag"
    SLACKBOT_EMOJI: "Tag"
    SLACKBOT_FILES: "Tag"
    SLACKBOT_FILES_COMMENTS: "Tag"
    SLACKBOT_FILES_REMOTE: "Tag"
    SLACKBOT_IMPORTANT: "Tag"
    SLACKBOT_MIGRATION: "Tag"
    SLACKBOT_OAUTH: "Tag"
    SLACKBOT_OAUTH_V2: "Tag"
    SLACKBOT_PINS: "Tag"
    SLACKBOT_REACTIONS: "Tag"
    SLACKBOT_REMINDERS: "Tag"
    SLACKBOT_RTM: "Tag"
    SLACKBOT_SEARCH: "Tag"
    SLACKBOT_STARS: "Tag"
    SLACKBOT_TEAM: "Tag"
    SLACKBOT_TEAM_PROFILE: "Tag"
    SLACKBOT_USERGROUPS: "Tag"
    SLACKBOT_USERGROUPS_USERS: "Tag"
    SLACKBOT_USERS: "Tag"
    SLACKBOT_USERS_PROFILE: "Tag"
    SLACKBOT_VIEWS: "Tag"
    SLACKBOT_WORKFLOWS: "Tag"
    SLACK_ADMIN: "Tag"
    SLACK_ADMIN_APPS: "Tag"
    SLACK_ADMIN_APPS_APPROVED: "Tag"
    SLACK_ADMIN_APPS_REQUESTS: "Tag"
    SLACK_ADMIN_APPS_RESTRICTED: "Tag"
    SLACK_ADMIN_CONVERSATIONS: "Tag"
    SLACK_ADMIN_CONVERSATIONS_EKM: "Tag"
    SLACK_ADMIN_CONVERSATIONS_RESTRICTACCESS: "Tag"
    SLACK_ADMIN_EMOJI: "Tag"
    SLACK_ADMIN_INVITEREQUESTS: "Tag"
    SLACK_ADMIN_INVITEREQUESTS_APPROVED: "Tag"
    SLACK_ADMIN_INVITEREQUESTS_DENIED: "Tag"
    SLACK_ADMIN_TEAMS: "Tag"
    SLACK_ADMIN_TEAMS_ADMINS: "Tag"
    SLACK_ADMIN_TEAMS_OWNERS: "Tag"
    SLACK_ADMIN_TEAMS_SETTINGS: "Tag"
    SLACK_ADMIN_USERGROUPS: "Tag"
    SLACK_ADMIN_USERS: "Tag"
    SLACK_ADMIN_USERS_SESSION: "Tag"
    SLACK_API: "Tag"
    SLACK_APPS: "Tag"
    SLACK_APPS_EVENT_AUTHORIZATIONS: "Tag"
    SLACK_APPS_PERMISSIONS: "Tag"
    SLACK_APPS_PERMISSIONS_RESOURCES: "Tag"
    SLACK_APPS_PERMISSIONS_SCOPES: "Tag"
    SLACK_APPS_PERMISSIONS_USERS: "Tag"
    SLACK_AUTH: "Tag"
    SLACK_BOTS: "Tag"
    SLACK_CALLS: "Tag"
    SLACK_CALLS_PARTICIPANTS: "Tag"
    SLACK_CHAT: "Tag"
    SLACK_CHAT_SCHEDULEDMESSAGES: "Tag"
    SLACK_CONVERSATIONS: "Tag"
    SLACK_DIALOG: "Tag"
    SLACK_DND: "Tag"
    SLACK_EMOJI: "Tag"
    SLACK_FILES: "Tag"
    SLACK_FILES_COMMENTS: "Tag"
    SLACK_FILES_REMOTE: "Tag"
    SLACK_IMPORTANT: "Tag"
    SLACK_MIGRATION: "Tag"
    SLACK_OAUTH: "Tag"
    SLACK_OAUTH_V2: "Tag"
    SLACK_PINS: "Tag"
    SLACK_REACTIONS: "Tag"
    SLACK_REMINDERS: "Tag"
    SLACK_RTM: "Tag"
    SLACK_SEARCH: "Tag"
    SLACK_STARS: "Tag"
    SLACK_TEAM: "Tag"
    SLACK_TEAM_PROFILE: "Tag"
    SLACK_USERGROUPS: "Tag"
    SLACK_USERGROUPS_USERS: "Tag"
    SLACK_USERS: "Tag"
    SLACK_USERS_PROFILE: "Tag"
    SLACK_VIEWS: "Tag"
    SLACK_WORKFLOWS: "Tag"
    SPOTIFY_ALBUMS: "Tag"
    SPOTIFY_ARTISTS: "Tag"
    SPOTIFY_AUDIOBOOKS: "Tag"
    SPOTIFY_CATEGORIES: "Tag"
    SPOTIFY_CHAPTERS: "Tag"
    SPOTIFY_EPISODES: "Tag"
    SPOTIFY_GENRES: "Tag"
    SPOTIFY_IMPORTANT: "Tag"
    SPOTIFY_LIBRARY: "Tag"
    SPOTIFY_MARKETS: "Tag"
    SPOTIFY_PLAYER: "Tag"
    SPOTIFY_PLAYLISTS: "Tag"
    SPOTIFY_SEARCH: "Tag"
    SPOTIFY_SHOWS: "Tag"
    SPOTIFY_TRACKS: "Tag"
    SPOTIFY_USERS: "Tag"
    STRIPE_CUSTOMER: "Tag"
    STRIPE_CUSTOMERS: "Tag"
    STRIPE_IMPORTANT: "Tag"
    STRIPE_INVOICES: "Tag"
    STRIPE_ONBOARDING: "Tag"
    STRIPE_PAYMENTS: "Tag"
    STRIPE_PRICES: "Tag"
    STRIPE_PRODUCTS: "Tag"
    STRIPE_SUBSCRIPTIONS: "Tag"
    SUPABASE_AUTH: "Tag"
    SUPABASE_DATABASE: "Tag"
    SUPABASE_DOMAINS: "Tag"
    SUPABASE_EDGE_FUNCTIONS: "Tag"
    SUPABASE_ENVIRONMENTS: "Tag"
    SUPABASE_IMPORTANT: "Tag"
    SUPABASE_OAUTH: "Tag"
    SUPABASE_ORGANIZATIONS: "Tag"
    SUPABASE_PROJECTS: "Tag"
    SUPABASE_REST: "Tag"
    SUPABASE_SECRETS: "Tag"
    SUPABASE_STORAGE: "Tag"
    TRELLO_ACTION: "Tag"
    TRELLO_BATCH: "Tag"
    TRELLO_BOARD: "Tag"
    TRELLO_CARD: "Tag"
    TRELLO_CHECKLIST: "Tag"
    TRELLO_IMPORTANT: "Tag"
    TRELLO_LABEL: "Tag"
    TRELLO_LIST: "Tag"
    TRELLO_MEMBER: "Tag"
    TRELLO_NOTIFICATION: "Tag"
    TRELLO_ORGANIZATION: "Tag"
    TRELLO_SEARCH: "Tag"
    TRELLO_SESSION: "Tag"
    TRELLO_TOKEN: "Tag"
    TRELLO_TYPE: "Tag"
    TRELLO_WEBHOOK: "Tag"
    TWITTER_BOOKMARKS: "Tag"
    TWITTER_COMPLIANCE: "Tag"
    TWITTER_DIRECT_MESSAGES: "Tag"
    TWITTER_GENERAL: "Tag"
    TWITTER_IMPORTANT: "Tag"
    TWITTER_LISTS: "Tag"
    TWITTER_SPACES: "Tag"
    TWITTER_TWEETS: "Tag"
    TWITTER_USAGE: "Tag"
    TWITTER_USERS: "Tag"
    WRIKE_ACCOUNTS: "Tag"
    WRIKE_CONTACTS: "Tag"
    WRIKE_CUSTOM_FIELDS: "Tag"
    WRIKE_FOLDER_PROJECTS: "Tag"
    WRIKE_GROUPS: "Tag"
    WRIKE_IMPORTANT: "Tag"
    WRIKE_INVITATIONS: "Tag"
    WRIKE_TASKS: "Tag"
    WRIKE_USERS: "Tag"
    WRIKE_WORKFLOWS: "Tag"
    ZOHO_CONVERT: "Tag"
    ZOHO_CREATE: "Tag"
    ZOHO_FETCH: "Tag"
    ZOHO_LEADS: "Tag"
    ZOHO_RECORDS: "Tag"
    ZOHO_RELATEDRECORDS: "Tag"
    ZOHO_TAGS: "Tag"
    ZOHO_UPDATE: "Tag"
    ZOHO_ZOHOCRM: "Tag"
    ZOOM_ARCHIVING: "Tag"
    ZOOM_CLOUD_RECORDING: "Tag"
    ZOOM_DEVICES: "Tag"
    ZOOM_H323_DEVICES: "Tag"
    ZOOM_IMPORTANT: "Tag"
    ZOOM_MEETINGS: "Tag"
    ZOOM_PAC: "Tag"
    ZOOM_REPORTS: "Tag"
    ZOOM_SIP_PHONE: "Tag"
    ZOOM_TRACKING_FIELD: "Tag"
    ZOOM_TSP: "Tag"
    ZOOM_WEBINARS: "Tag"



"""
Helper utilities.
"""

import sys

from .enums import get_enum_key


def blue(message: str) -> str:
    return f"\033[36m{message}\033[0m"


def green(message: str) -> str:
    return f"\033[32m{message}\033[0m"


help_msg_already_printed = False


def help_msg() -> str:
    global help_msg_already_printed
    if help_msg_already_printed:
        return ""

    help_msg_already_printed = True
    if sys.stdout.isatty():
        return (
            blue("Give Feedback / Get Help:\n")
            + blue("    On GitHub: ")
            + "https://github.com/ComposioHQ/composio/issues/new\n"
            + blue("    On Discord: ")
            + "https://dub.composio.dev/discord\n"
            + blue("    On Email: ")
            + "tech@composio.dev\n"
            + blue("    Talk to us on Intercom: ")
            + "https://composio.dev\n"
            + blue("    Book a call with us: ")
            + "https://calendly.com/soham-composio/30min\n"
            + "If you need to debug this error, "
            + green("set `COMPOSIO_LOGGING_LEVEL=debug`")
            + "."
        )

    return (
        "Give Feedback / Get Help:\n"
        "    On GitHub: https://github.com/ComposioHQ/composio/issues/new\n"
        "    On Discord: https://dub.composio.dev/discord\n"
        "    On Email: tech@composio.dev\n"
        "    Talk to us on Intercom: https://composio.dev\n"
        "    Book a call with us: https://calendly.com/soham-composio/30min\n"
        "If you need to debug this error, set `COMPOSIO_LOGGING_LEVEL=debug`."
    )



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetTaskSTimeInStatusRequest(BaseModel):
    """Request schema for `GetTaskSTimeInStatus`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )
    custom_task_ids: t.Optional[bool] = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: t.Optional[int] = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )


class GetTaskSTimeInStatusResponse(BaseModel):
    """Response schema for `GetTaskSTimeInStatus`"""

    data: t.Dict[str, t.Any]


class GetTaskSTimeInStatus(OpenAPIAction):
    """
    View how long a task has been in each status. The Total time in Status ClickApp
    must first be enabled by the Workspace owner or an admin.
    """

    _tags = ["Tasks"]
    _display_name = "get_task_s_time_in_status"
    _request_schema = GetTaskSTimeInStatusRequest
    _response_schema = GetTaskSTimeInStatusResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/time_in_status"
    method = "get"
    operation_id = "Tasks_getTimeInStatus"
    action_identifier = "/task/{task_id}/time_in_status_get"

    path_params = {"task_id": "task_id"}
    query_params = {"custom_task_ids": "custom_task_ids", "team_id": "team_id"}
    header_params = {}
    request_params = {}

    aliases = {}



import typing as t

from composio.tools.base.local import LocalAction, LocalTool

from .actions import Crawl, Scrape


class Spidertool(LocalTool, autoload=True):
    """Spider Tools"""

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/spidertool.png"

    @classmethod
    def actions(cls) -> list[t.Type[LocalAction]]:
        return [Scrape, Crawl]



import typing as t

from composio.client.enums.action import Action
from composio.client.enums.enum import Enum, EnumGenerator

from .base import AppData

_APP_CACHE: t.Dict[str, "App"] = {}

class App(Enum[AppData], metaclass=EnumGenerator):
    cache_folder = "apps"
    cache = _APP_CACHE
    storage = AppData

    def load_from_runtime(self) -> t.Optional[AppData]: ...
    def fetch_and_cache(self) -> t.Optional[AppData]: ...
    @property
    def name(self) -> str: ...
    @property
    def is_local(self) -> bool: ...
    def get_actions(
        self, tags: t.Optional[t.List[str]] = None
    ) -> t.Iterator[Action]: ...
    ABLY: "App"
    ACCELO: "App"
    ACTIVE_CAMPAIGN: "App"
    ADOBE: "App"
    AERO_WORKFLOW: "App"
    AFFINITY: "App"
    AGENCYZOOM: "App"
    AHREFS: "App"
    AIRTABLE: "App"
    ALCHEMY: "App"
    ALTOVIZ: "App"
    AMAZON: "App"
    AMCARDS: "App"
    AMPLITUDE: "App"
    ANTHROPIC: "App"
    APOLLO: "App"
    APPDRAG: "App"
    APPSFLYER: "App"
    ASANA: "App"
    ASHBY: "App"
    ATLASSIAN: "App"
    ATTIO: "App"
    AUTH0: "App"
    AXONAUT: "App"
    BAMBOOHR: "App"
    BANNERBEAR: "App"
    BASEROW: "App"
    BATTLENET: "App"
    BEEMINDER: "App"
    BITBUCKET: "App"
    BITWARDEN: "App"
    BLACKBAUD: "App"
    BLACKBOARD: "App"
    BOLDSIGN: "App"
    BOTBABA: "App"
    BOX: "App"
    BRAINTREE: "App"
    BRANDFETCH: "App"
    BREEZY_HR: "App"
    BREVO: "App"
    BREX: "App"
    BREX_STAGING: "App"
    BRIGHTPEARL: "App"
    BROWSEAI: "App"
    BROWSERBASE_TOOL: "App"
    BROWSERHUB: "App"
    BROWSER_TOOL: "App"
    BUBBLE: "App"
    CAL: "App"
    CALENDLY: "App"
    CANVA: "App"
    CANVAS: "App"
    CHATWORK: "App"
    CHMEETINGS: "App"
    CLICKUP: "App"
    CLOSE: "App"
    CLOUDFLARE: "App"
    CODEINTERPRETER: "App"
    CODE_ANALYSIS_TOOL: "App"
    CODE_FORMAT_TOOL: "App"
    COINMARKETCAL: "App"
    COMPOSIO: "App"
    CONTENTFUL: "App"
    CUSTOMER_IO: "App"
    DAILYBOT: "App"
    DATADOG: "App"
    DATAGMA: "App"
    DATAROBOT: "App"
    DEEL: "App"
    DEMIO: "App"
    DIGICERT: "App"
    DISCORD: "App"
    DISCORDBOT: "App"
    DOCMOSIS: "App"
    DOCUSIGN: "App"
    DROPBOX: "App"
    DROPBOX_SIGN: "App"
    DYNAMICS365: "App"
    ECHTPOST: "App"
    ELEVENLABS: "App"
    EMBED_TOOL: "App"
    ENTELLIGENCE: "App"
    EPIC_GAMES: "App"
    EVENTBRITE: "App"
    EXA: "App"
    EXIST: "App"
    FACEBOOK: "App"
    FACTORIAL: "App"
    FIGMA: "App"
    FILETOOL: "App"
    FINAGE: "App"
    FIRECRAWL: "App"
    FITBIT: "App"
    FLUTTERWAVE: "App"
    FOMO: "App"
    FORMCARRY: "App"
    FORMSITE: "App"
    FRESHBOOKS: "App"
    FRESHDESK: "App"
    FRONT: "App"
    GIT: "App"
    GITHUB: "App"
    GMAIL: "App"
    GOOGLEBIGQUERY: "App"
    GOOGLECALENDAR: "App"
    GOOGLEDOCS: "App"
    GOOGLEDRIVE: "App"
    GOOGLEMEET: "App"
    GOOGLEPHOTOS: "App"
    GOOGLESHEETS: "App"
    GOOGLETASKS: "App"
    GOOGLE_ANALYTICS: "App"
    GOOGLE_MAPS: "App"
    GORGIAS: "App"
    GO_TO_WEBINAR: "App"
    GREPTILE: "App"
    GUMROAD: "App"
    GURU: "App"
    HACKERNEWS: "App"
    HACKERRANK_WORK: "App"
    HARVEST: "App"
    HELCIM: "App"
    HEYGEN: "App"
    HIGHLEVEL: "App"
    HISTORY_FETCHER: "App"
    HUBSPOT: "App"
    HUMANLOOP: "App"
    ICIMS_TALENT_CLOUD: "App"
    IDEA_SCALE: "App"
    IMAGE_ANALYSER: "App"
    INDUCED_AI: "App"
    INTERCOM: "App"
    INTERZOID: "App"
    JIRA: "App"
    JUNGLESCOUT: "App"
    KEAP: "App"
    KLAVIYO: "App"
    KLIPFOLIO: "App"
    LASTPASS: "App"
    LAUNCH_DARKLY: "App"
    LEVER: "App"
    LEVER_SANDBOX: "App"
    LEXOFFICE: "App"
    LINEAR: "App"
    LINKEDIN: "App"
    LINKHUT: "App"
    LISTENNOTES: "App"
    LMNT: "App"
    MAILCHIMP: "App"
    MAILERLITE: "App"
    MAINTAINX: "App"
    MATHEMATICAL: "App"
    MBOUM: "App"
    MEM0: "App"
    METATEXTAI: "App"
    MICROSOFT_CLARITY: "App"
    MICROSOFT_TEAMS: "App"
    MICROSOFT_TENANT: "App"
    MIRO: "App"
    MIXPANEL: "App"
    MOCEAN: "App"
    MONDAY: "App"
    MOPINION: "App"
    MORE_TREES: "App"
    MOXIE: "App"
    MOZ: "App"
    MULTIONAI: "App"
    MURAL: "App"
    NCSCALE: "App"
    NEON: "App"
    NETSUITE: "App"
    NGROK: "App"
    NOTION: "App"
    ONCEHUB: "App"
    ONE_DRIVE: "App"
    OUTLOOK: "App"
    PAGERDUTY: "App"
    PANDADOC: "App"
    PEOPLEDATALABS: "App"
    PERPLEXITYAI: "App"
    PIGGY: "App"
    PIPEDRIVE: "App"
    PLACEKEY: "App"
    POSTHOG: "App"
    PRECORO: "App"
    PRINTNODE: "App"
    PROCESS_STREET: "App"
    PRODUCTBOARD: "App"
    QUALAROO: "App"
    RAFFLYS: "App"
    RAGTOOL: "App"
    RAVENSEOTOOLS: "App"
    REDDIT: "App"
    RETELLAI: "App"
    RING_CENTRAL: "App"
    RIPPLING: "App"
    ROCKETLANE: "App"
    ROCKET_REACH: "App"
    SAGE: "App"
    SALESFORCE: "App"
    SCREENSHOTONE: "App"
    SEISMIC: "App"
    SENDGRID: "App"
    SERPAPI: "App"
    SERVICEM8: "App"
    SHARE_POINT: "App"
    SHELLTOOL: "App"
    SHOPIFY: "App"
    SHORTCUT: "App"
    SIMPLESAT: "App"
    SLACK: "App"
    SLACKBOT: "App"
    SMARTRECRUITERS: "App"
    SMUGMUG: "App"
    SNOWFLAKE: "App"
    SPIDERTOOL: "App"
    SPOTIFY: "App"
    SQLTOOL: "App"
    SQUARE: "App"
    STACK_EXCHANGE: "App"
    STRIPE: "App"
    SUPABASE: "App"
    SURVEY_MONKEY: "App"
    TAPFORM: "App"
    TAVILY: "App"
    TERMINUS: "App"
    TEXTRAZOR: "App"
    TIMECAMP: "App"
    TIMEKIT: "App"
    TIMELY: "App"
    TINYPNG: "App"
    TINYURL: "App"
    TISANE: "App"
    TODOIST: "App"
    TONEDEN: "App"
    TRELLO: "App"
    TWITCH: "App"
    TWITTER: "App"
    TWITTER_MEDIA: "App"
    VENLY: "App"
    VERO: "App"
    VISME: "App"
    WABOXAPP: "App"
    WAKATIME: "App"
    WAVE_ACCOUNTING: "App"
    WEATHERMAP: "App"
    WEBEX: "App"
    WEBFLOW: "App"
    WEBTOOL: "App"
    WIZ: "App"
    WORKIOM: "App"
    WORKSPACE_TOOL: "App"
    WRIKE: "App"
    XERO: "App"
    YANDEX: "App"
    YNAB: "App"
    YOUSEARCH: "App"
    YOUTUBE: "App"
    ZENDESK: "App"
    ZENROWS: "App"
    ZENSERP: "App"
    ZEPTOOL: "App"
    ZOHO: "App"
    ZOHO_BIGIN: "App"
    ZOHO_BOOKS: "App"
    ZOHO_DESK: "App"
    ZOHO_INVENTORY: "App"
    ZOHO_INVOICE: "App"
    ZOHO_MAIL: "App"
    ZOOM: "App"
    ZOOMINFO: "App"



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class TrackTimeRequest(BaseModel):
    """Request schema for `TrackTime`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="",
    )
    custom_task_ids: t.Optional[bool] = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: t.Optional[int] = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`.   For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )
    start: int = Field(
        default=...,
        alias="start",
        description="Start",
    )
    end: int = Field(
        default=...,
        alias="end",
        description="End",
    )
    time: int = Field(
        default=...,
        alias="time",
        description="Time",
    )


class TrackTimeResponse(BaseModel):
    """Response schema for `TrackTime`"""

    data: t.Dict[str, t.Any]


class TrackTime(OpenAPIAction):
    """
    ***Note:** This is a legacy time tracking endpoint. We recommend using the
    Time Tracking API endpoints to manage time entries.*
    """

    _tags = ["Time Tracking (Legacy)"]
    _display_name = "track_time"
    _request_schema = TrackTimeRequest
    _response_schema = TrackTimeResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/time"
    method = "post"
    operation_id = "TimeTrackingLegacy_recordTimeForTask"
    action_identifier = "/task/{task_id}/time_post"

    path_params = {"task_id": "task_id"}
    query_params = {"custom_task_ids": "custom_task_ids", "team_id": "team_id"}
    header_params = {}
    request_params = {
        "start": {"__alias": "start"},
        "end": {"__alias": "end"},
        "time": {"__alias": "time"},
    }

    aliases = {}



import typing as t

from composio.client.enums.enum import Enum, EnumGenerator

from .base import TagData


_TAG_CACHE: t.Dict[str, "Tag"] = {}


class Tag(Enum[TagData], metaclass=EnumGenerator):
    cache_folder = "tags"
    cache = _TAG_CACHE
    storage = TagData

    def fetch_and_cache(self) -> t.Optional[TagData]:
        # Tags are only fetched at startup right now.
        return None

    def load_from_runtime(self) -> t.Optional[TagData]:
        # Tags are not used at runtime right now.
        return None

    # TODO: why is it value in tag but name in trigger
    @property
    def value(self) -> str:
        return self.load().value

    @property
    def app(self) -> str:
        return self.load().app



"""Base abstractions."""

import hashlib
import inspect
import json
import textwrap
import typing as t
from abc import abstractmethod
from pathlib import Path

import inflection
import jsonref
import pydantic
from pydantic import BaseModel, Field

from composio.client.enums import Action as ActionEnum
from composio.client.enums.base import DEPRECATED_MARKER
from composio.exceptions import ComposioSDKError
from composio.utils.logging import WithLogger
from composio.utils.pydantic import parse_pydantic_error


GroupID = t.Literal["runtime", "local", "api"]
ModelType = t.TypeVar("ModelType")
ActionResponse = t.TypeVar("ActionResponse")
ActionRequest = t.TypeVar("ActionRequest")
Loadable = t.TypeVar("Loadable")
ToolRegistry = t.Dict[GroupID, t.Dict[str, "Tool"]]
ActionsRegistry = t.Dict[GroupID, t.Dict[str, "Action"]]
# TODO: create a Trigger type for this
TriggersRegistry = t.Dict[GroupID, t.Dict[str, t.Any]]

tool_registry: ToolRegistry = {"runtime": {}, "local": {}, "api": {}}
action_registry: ActionsRegistry = {"runtime": {}, "local": {}, "api": {}}
trigger_registry: TriggersRegistry = {"runtime": {}, "local": {}, "api": {}}


def remove_json_ref(data: t.Dict) -> t.Dict:
    return json.loads(
        jsonref.dumps(
            jsonref.replace_refs(
                obj=data,
                lazy_load=False,
                merge_props=True,
            ),
            indent=2,
        )
    )


def generate_app_id(name: str) -> str:
    # Generate a 32-character hash using MD5
    hash_string = hashlib.md5(name.encode()).hexdigest()
    # Insert hyphens at the specified positions
    return "-".join(
        (
            hash_string[:8],
            hash_string[8:12],
            hash_string[12:16],
            hash_string[16:20],
            hash_string[20:],
        )
    )


class InvalidClassDefinition(ComposioSDKError):
    """Raise when a class is not defined properly."""


class ExecuteResponse(BaseModel):
    """Execute action response."""


class _Attributes:
    name: str
    """Name representation."""

    enum: str
    """Enum key."""

    display_name: str
    """Display compatible name."""

    description: str
    """Description string."""

    logo: str
    """URL for the resource logo."""


class _Request(t.Generic[ModelType]):
    """Request util."""

    def __init__(self, model: t.Type[ModelType]) -> None:
        """Initialize request model."""
        self.model = model

    def schema(self) -> t.Dict:
        """Build request schema."""
        request = t.cast(t.Type[BaseModel], self.model).model_json_schema(by_alias=True)
        request = remove_json_ref(request)
        if "$defs" in request:
            del request["$defs"]

        properties = request.get("properties", {})
        for prop in properties.values():
            if prop.get("file_readable", False):
                prop["oneOf"] = [
                    {
                        "type": prop.get("type"),
                        "description": prop.get("description", ""),
                    },
                    {
                        "type": "string",
                        "format": "file-path",
                        "description": f"File path to {prop.get('description', '')}",
                    },
                ]
                del prop["type"]  # Remove original type to avoid conflict in oneOf
                continue

            if "allOf" in prop and len(prop["allOf"]) == 1:
                (schema,) = prop.pop("allOf")
                prop.update(schema)
                if "enum" in schema:
                    prop[
                        "description"
                    ] += f" Note: choose value only from following options - {prop['enum']}"

            if "anyOf" in prop:
                typedef, *_ = [
                    td for td in prop["anyOf"] if td.get("type", "null") != "null"
                ]
                prop["type"] = typedef["type"]

        request["properties"] = properties
        return request

    def parse(self, request: t.Dict) -> ModelType:
        """Parse request."""
        try:
            return self.model(**request)
        except pydantic.ValidationError as e:
            raise ValueError(parse_pydantic_error(e)) from e


class _Response(t.Generic[ModelType]):
    """Response util."""

    def __init__(self, model: t.Type[ModelType]) -> None:
        """Initialize request model."""
        self.model = model
        self.wrapper = self.wrap(model=model)

    @classmethod
    def wrap(cls, model: t.Type[ModelType]) -> t.Type[BaseModel]:
        class wrapper(model):  # type: ignore
            successful: bool = Field(
                ...,
                description="Whether or not the action execution was successful or not",
            )
            error: t.Optional[str] = Field(
                None,
                description="Error if any occurred during the execution of the action",
            )

        return t.cast(t.Type[BaseModel], wrapper)

    def schema(self) -> t.Dict:
        """Build request schema."""
        schema = self.wrapper.model_json_schema(by_alias=True)
        schema = remove_json_ref(schema)
        if "$defs" in schema:
            del schema["$defs"]

        properties = schema.get("properties", {})
        for prop in properties.values():
            if prop.get("file_readable", False):
                prop["oneOf"] = [
                    {
                        "type": prop.get("type"),
                        "description": prop.get("description", ""),
                    },
                    {
                        "type": "string",
                        "format": "file-path",
                        "description": f"File path to {prop.get('description', '')}",
                    },
                ]
                del prop["type"]  # Remove original type to avoid conflict in oneOf
                continue

            if "allOf" in prop and len(prop["allOf"]) == 1:
                (schema,) = prop.pop("allOf")
                prop.update(schema)
                if "enum" in schema:
                    prop[
                        "description"
                    ] += f" Note: choose value only from following options - {prop['enum']}"

        schema["properties"] = properties
        schema["title"] = self.model.__name__
        return remove_json_ref(schema)


class ActionBuilder:
    @staticmethod
    def get_generics(obj: t.Type["Action"]) -> t.Tuple[t.Any, t.Any]:
        for base in getattr(obj, "__orig_bases__"):
            args = t.get_args(base)
            if len(args) == 2:
                return args  # type: ignore
        raise ValueError("No type generics found")

    @classmethod
    def set_generics(cls, name: str, obj: t.Type["Action"]) -> None:
        try:
            request, response = cls.get_generics(obj=obj)
            if request == ActionRequest or response == ActionResponse:
                raise ValueError(f"Invalid type generics, ({request}, {response})")
        except ValueError as e:
            raise InvalidClassDefinition(
                "Invalid action class definition, please define your class "
                "using request and response type generics; "
                f"class {name}(Action[RequestModel, ResponseModel])"
            ) from e

        setattr(obj, "request", _Request(request))
        setattr(obj, "response", _Response(response))

    @staticmethod
    def validate(name: str, obj: t.Type["Action"]) -> None:
        if getattr(getattr(obj, "execute"), "__isabstractmethod__", False):
            raise InvalidClassDefinition(f"Please implement {name}.execute")

    @classmethod
    def set_metadata(cls, obj: t.Type["Action"]) -> None:
        setattr(obj, "file", getattr(obj, "file", Path(inspect.getfile(obj))))
        setattr(obj, "name", getattr(obj, "name", inflection.underscore(obj.__name__)))
        setattr(
            obj,
            "enum",
            getattr(obj, "enum", inflection.underscore(obj.__name__).upper()),
        )
        setattr(
            obj,
            "display_name",
            getattr(
                obj,
                "display_name",
                inflection.humanize(inflection.underscore(obj.__name__)),
            ),
        )
        setattr(
            obj,
            "description",
            cls._get_description(obj),
        )
        description, *_ = obj.description.split(DEPRECATED_MARKER, maxsplit=1)
        if len(description) > 1024:
            raise InvalidClassDefinition(
                f"Description for action `{obj.__name__}` contains more than 1024 characters"
            )

    @staticmethod
    def _get_description(obj) -> str:
        description = t.cast(
            str,
            (
                (obj.__doc__ if obj.__doc__ else obj.display_name)
                .replace("\n    ", " ")
                .strip()
            ),
        )
        description, separator, enum = description.partition(DEPRECATED_MARKER)
        return inflection.titleize(description) + separator + enum


class ActionMeta(type):
    """Action metaclass."""

    def __init__(  # pylint: disable=unused-argument, self-cls-assignment
        cls,
        name: str,
        bases: t.Tuple,
        dict_: t.Dict,
        abs: bool = False,
    ) -> None:
        """Initialize action class."""
        if abs or name == "Action":
            return

        cls = t.cast(t.Type["Action"], cls)
        ActionBuilder.validate(name=name, obj=cls)
        ActionBuilder.set_generics(name=name, obj=cls)
        ActionBuilder.set_metadata(obj=cls)


class Action(
    WithLogger,
    _Attributes,
    t.Generic[ActionRequest, ActionResponse],
    metaclass=ActionMeta,
):
    """Action abstraction."""

    _tags: t.Optional[t.List[str]] = None

    _schema: t.Optional[t.Dict] = None

    tool: str
    """Toolname."""

    request: _Request[ActionRequest]
    """Request helper."""

    response: _Response[ActionResponse]
    """Response helper."""

    file: str
    """Path to the file containing the action"""

    requires: t.Optional[t.List[str]] = None
    """List of dependencies required to run this action."""

    no_auth: bool = False
    """If set `True`, the action does not require a connected account."""

    def __init_subclass__(cls, abs: bool = False) -> None:
        """Initialize subclas."""

    @classmethod
    def tags(cls) -> t.List[str]:
        """Tags for the given action."""
        return cls._tags or []

    @classmethod
    def _generate_schema(cls) -> None:
        """Generate action schema."""
        cls._schema = {
            "name": cls.name,
            "enum": cls.enum,
            "appName": cls.tool,
            "appId": generate_app_id(cls.tool),
            "tags": cls.tags(),
            "response": cls.response.schema(),
            "parameters": cls.request.schema(),
            "displayName": cls.display_name,
            "description": cls.description,
        }

    @classmethod
    def schema(cls) -> t.Dict:
        """Action schema."""
        if cls._schema is None:
            cls._generate_schema()
        return cls._schema  # type: ignore

    @abstractmethod
    def execute(
        self,
        request: ActionRequest,
        metadata: t.Dict,
    ) -> ActionResponse:
        """Execute the action."""


class ToolBuilder:
    @staticmethod
    def validate(obj: t.Type["Tool"], name: str, methods: t.Tuple[str, ...]) -> None:
        for method in methods:
            if getattr(getattr(obj, method), "__isabstractmethod__", False):
                raise InvalidClassDefinition(f"Please implement {name}.{method}")

            if not inspect.ismethod(getattr(obj, method)):
                raise InvalidClassDefinition(
                    f"Please implement {name}.{method} as class method"
                )

    @classmethod
    def set_metadata(cls, obj: t.Type["Tool"]) -> None:
        setattr(obj, "file", Path(inspect.getfile(obj)))
        setattr(obj, "gid", getattr(obj, "gid", "local"))
        setattr(obj, "name", getattr(obj, "name", inflection.underscore(obj.__name__)))
        setattr(obj, "enum", getattr(obj, "enum", obj.name).upper())
        setattr(
            obj,
            "display_name",
            getattr(
                obj,
                "display_name",
                inflection.humanize(inflection.underscore(obj.__name__)),
            ),
        )
        setattr(obj, "description", cls._get_description(obj=obj))
        setattr(obj, "_actions", getattr(obj, "_actions", {}))
        setattr(obj, "_triggers", getattr(obj, "_triggers", {}))

    @staticmethod
    def setup_children(obj: t.Type["Tool"], no_auth: bool = False) -> None:
        if obj.gid not in action_registry:
            action_registry[obj.gid] = {}

        for action in obj.actions():
            action.tool = obj.name
            action.enum = f"{obj.enum}_{action.name.upper()}"
            action.no_auth = no_auth
            if obj.requires is not None:
                action.requires = list(set(obj.requires + (action.requires or [])))
            obj._actions[action.enum] = action  # pylint: disable=protected-access
            action_registry[obj.gid][action.enum] = action  # type: ignore

            if hasattr(obj, "logo"):
                setattr(action, "logo", getattr(obj, "logo"))

        if not hasattr(obj, "triggers"):
            return

        if obj.gid not in trigger_registry:
            trigger_registry[obj.gid] = {}

        for trigger in obj.triggers():  # type: ignore
            trigger.tool = obj.name
            trigger.enum = f"{obj.enum}_{trigger.name.upper()}"
            obj._triggers[trigger.enum] = trigger  # type: ignore  # pylint: disable=protected-access
            trigger_registry[obj.gid][trigger.enum] = trigger  # type: ignore

            if hasattr(obj, "logo"):
                setattr(trigger, "logo", getattr(obj, "logo"))

    @staticmethod
    def _get_description(obj) -> str:
        return " ".join(
            line
            for line in textwrap.dedent(
                (obj.__doc__ if obj.__doc__ else obj.display_name)
            )
            .strip()
            .splitlines()
            if line
        )


class Tool(WithLogger, _Attributes):
    """Tool abstraction."""

    gid: GroupID
    """Group ID for this tool."""

    file: Path
    """Path to module file."""

    name: str
    """Tool name."""

    requires: t.Optional[t.List[str]] = None
    """List of dependencies required for running this tool."""

    _schema: t.Optional[t.Dict] = None
    """Schema for the app."""

    _actions: t.Dict[str, t.Type[Action]]
    """Actions container"""

    def __init_subclass__(cls, autoload: bool = False) -> None:
        """Initialize a tool class."""

    @classmethod
    def get(cls, enum: ActionEnum) -> t.Type[Action]:
        """Returns the"""
        return cls._actions[enum.slug]

    @classmethod
    @abstractmethod
    def actions(cls) -> t.List[t.Type[t.Any]]:
        """Get collection of actions for the tool."""

    @classmethod
    def _generate_schema(cls) -> None:
        """Generate schema for the app."""
        cls._schema = {
            "name": cls.name,
            "displayName": cls.display_name,
            "metaData": {
                "toolName": cls.name,
                "groupId": cls.gid,
                "displayName": cls.display_name,
                "description": cls.description,
                "toolPath": str(cls.file),
            },
            "integration": {},
            "description": cls.description,
            "actions": [action.schema() for action in cls.actions()],
        }

    @classmethod
    def schema(cls) -> t.Dict:
        """Get tool schema."""
        if cls._schema is None:
            cls._generate_schema()
        return cls._schema  # type: ignore

    def _load(self, loadable: t.Type[Loadable]) -> Loadable:
        """Load action class."""
        instance = loadable()
        return instance

    def execute(
        self,
        action: str,
        params: t.Dict,
        metadata: t.Optional[t.Dict] = None,
    ) -> t.Dict:
        """
        Execute the given action

        :param action: Name of the action.
        :param params: Execution parameters.
        :param metadata: A dictionary containing metadata for action.
        """
        raise NotImplementedError()

    @classmethod
    def register(cls: t.Type["Tool"]) -> None:
        """Register given tool to the registry."""
        # TODO(Viraj): Check if the tool is already registered
        if cls.gid not in tool_registry:
            tool_registry[cls.gid] = {}
        tool_registry[cls.gid][cls.enum] = cls()



"""
Apps manager for Composio SDK.

Usage:
    composio apps [command] [options]
"""

import ast
import os.path
import shutil

import click

from composio import constants
from composio.cli.context import Context, pass_context
from composio.cli.utils.decorators import handle_exceptions
from composio.cli.utils.helpfulcmd import HelpfulCmdBase
from composio.client import Composio
from composio.client.enums.base import (
    ACTIONS_CACHE,
    APPS_CACHE,
    TAGS_CACHE,
    TRIGGERS_CACHE,
)
from composio.client.utils import update_actions, update_apps, update_triggers
from composio.core.cls.did_you_mean import DYMGroup
from composio.exceptions import ComposioSDKError


class AppsExamples(HelpfulCmdBase, DYMGroup):
    examples = [
        click.style("composio apps", fg="green")
        + click.style("                # List all apps\n", fg="black"),
        click.style("composio apps --enabled", fg="green")
        + click.style("      # List only enabled apps\n", fg="black"),
        click.style("composio apps generate-types", fg="green")
        + click.style(" # Update type stubs for enums\n", fg="black"),
    ]


@click.group(name="apps", invoke_without_command=True, cls=AppsExamples)
@click.help_option("--help", "-h", "-help")
@click.option(
    "--enabled",
    is_flag=True,
    default=False,
    help="Only show apps which are enabled",
)
@handle_exceptions()
@pass_context
def _apps(context: Context, enabled: bool = False) -> None:
    """List composio tools/apps which you have access to"""
    if context.click_ctx.invoked_subcommand:
        return

    apps = context.client.apps.get()
    if enabled:
        apps = [app for app in apps if app.enabled]
        context.console.print("[green]Showing apps which are enabled[/green]")
    else:
        context.console.print("[green]Showing all apps[/green]")

    for app in apps:
        context.console.print(f"• {app.key}")


@_apps.command(name="update")
@click.help_option("--help", "-h", "-help")
@handle_exceptions()
@pass_context
def _update(context: Context) -> None:
    """Deprecated, has no effect."""
    context.console.print(
        "[yellow]Warning:[/yellow] the 'apps update' command has been deprecated"
        " and has no effect.\n"
        "If you wish to update the local type stubs for auto-completion, use"
        " 'composio apps generate-types' instead."
    )


@_apps.command(name="generate-types")
@click.help_option("--help", "-h", "-help")
@handle_exceptions()
@pass_context
def _generate_types(context: Context) -> None:
    """Updates the local type stubs with the latest app data."""
    context.console.print("Fetching latest data from Composio API...")
    generate_type_stubs(context.client)
    context.console.print(
        "[green]Successfully updated type stubs for Apps, Actions, and Triggers[/green]"
    )


def generate_type_stub(enum_file: str, cache_folder: os.PathLike) -> None:
    # Get all enum filenames
    enum_names = sorted(os.listdir(cache_folder))

    # Get the enum class
    with open(enum_file, encoding="utf-8") as f:
        tree = ast.parse(f.read())

    enum_classes = [
        node
        for node in tree.body
        if isinstance(node, ast.ClassDef)
        and isinstance(node.bases[0], ast.Subscript)
        and isinstance(node.bases[0].value, ast.Name)
        and node.bases[0].value.id == "Enum"
    ]
    if not enum_classes:
        raise ComposioSDKError(
            "No Enum class found in the SDK source, please re-install `composio`."
        )
    enum_class = enum_classes[0]

    # Remove the bodies of all methods in the class, replace with ellipsis
    for node in enum_class.body:
        if isinstance(node, ast.FunctionDef):
            node.body = [ast.Expr(ast.Constant(...))]

    # Add all enum names as class attributes
    for enum_name in enum_names:
        enum_class.body.append(
            ast.AnnAssign(
                target=ast.Name(id=enum_name, ctx=ast.Store()),
                annotation=ast.Constant(value=f"{enum_class.name}"),
                value=None,
                simple=1,
            )
        )

    # Write the type stub
    with open(enum_file + "i", "w", encoding="utf-8") as f:
        f.write(ast.unparse(tree))


def generate_type_stubs(client: Composio) -> None:
    # Update local cache first
    for cache_folder in ["apps", "actions", "triggers", "tags"]:
        shutil.rmtree(
            constants.LOCAL_CACHE_DIRECTORY / cache_folder,
            ignore_errors=True,
        )
    apps = update_apps(client)
    update_actions(client, apps)
    update_triggers(client, apps)

    enums_folder = os.path.join(os.path.dirname(__file__), "..", "client", "enums")
    apps_enum = os.path.join(enums_folder, "app.py")
    actions_enum = os.path.join(enums_folder, "action.py")
    triggers_enum = os.path.join(enums_folder, "trigger.py")
    tags_enum = os.path.join(enums_folder, "tag.py")

    for enum_file, cache_folder_path in [
        (apps_enum, APPS_CACHE),
        (actions_enum, ACTIONS_CACHE),
        (triggers_enum, TRIGGERS_CACHE),
        (tags_enum, TAGS_CACHE),
    ]:
        generate_type_stub(enum_file, cache_folder_path)
